import numpy as np
import numpy.typing as npt


def dividAndConquer(
    arr1: npt.NDArray[np.integer | np.floating],
    arr2: npt.NDArray[np.integer | np.floating],
) -> tuple[npt.NDArray[np.floating], npt.NDArray[np.floating]]:
    """
    Pure numpy implementation of the original
    'divide and conquer fast intersection algorithm. Waithe D 2014'

    # Note
    2026-01-16 Thomas Bischof

    Replaces the following block from fib4.pyx. I tested this against both
    random data and real data, and obtained identical results.
    --------------------------------------------
    PicoQuant functionalities from FCS_viewer

    This file contains a fast implementation of an algorithm that is
    very important (yes I have no clue about the structure of pt3 files)
    for importing *.pt3 files: `dividAndConquer`.

    The code was written by
    Dr. Dominic Waithe
    Wolfson Imaging Centre.
    Weatherall Institute of Molecular Medicine.
    University of Oxford

    https://github.com/dwaithe/FCS_viewer

    See Also:
        The wrapper: `read_pt3_PicoQuant.py`
        The wrapped file: `read_pt3_PicoQuant_original_FCSViewer.py`.

    import cython
    cimport cython

    import numpy as np
    cimport numpy as np

    DTYPE = np.float64
    ctypedef np.float64_t DTYPE_t

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.nonecheck(False)
    def dividAndConquer(arr1b,arr2b,arrLength):
        \"\"\"divide and conquer fast intersection algorithm. Waithe D 2014\"\"\"

        cdef np.ndarray[DTYPE_t, ndim=1] arr1bool = np.zeros((arrLength-1))
        cdef np.ndarray[DTYPE_t, ndim=1] arr2bool = np.zeros((arrLength-1))
        cdef np.ndarray[DTYPE_t, ndim=1] arr1 = arr1b
        cdef np.ndarray[DTYPE_t, ndim=1] arr2 = arr2b

        cdef int arrLen
        arrLen = arrLength;
        cdef int i
        i = 0;
        cdef int j
        j = 0;

        while(i <arrLen-1 and j< arrLen-1):

            if(arr1[i] < arr2[j]):
              i+=1;
            elif(arr2[j] < arr1[i]):
              j+=1;
            elif (arr1[i] == arr2[j]):

              arr1bool[i] = 1;
              arr2bool[j] = 1;
              i+=1;

        return arr1bool,arr2bool
    """
    a1 = arr1[:-1]
    a2 = arr2[:-1]

    idx1 = np.searchsorted(a2, a1)
    idx1[idx1 == len(a2)] = len(a2) - 1
    mask1 = (a1 == a2[idx1]).astype(np.float64)

    idx2 = np.searchsorted(a1, a2)
    idx2[idx2 == len(a1)] = len(a1) - 1
    mask2 = (a2 == a1[idx2]).astype(np.float64)

    return mask1, mask2



"""correlator.com .sin files"""
import csv
import pathlib
import warnings

import numpy as np


class OpenSINError(BaseException):
    pass


def openSIN(path, filename=None):
    """Parse .sin files (correlator.com)"""
    path = pathlib.Path(path)
    if filename is not None:
        warnings.warn("Using `filename` is deprecated.", DeprecationWarning)
        path = path / filename
    filename = path.name

    with path.open() as fd:
        data = fd.readlines()

    for line in data:
        line = line.strip()
        if line.lower().startswith("mode"):
            mode = line.split("=")[1].strip().split()
            # Find out what kind of mode it is

            # The rationale is that when the mode
            # consists of single characters separated
            # by empty spaces, then we have integer mode.
            if len(mode) - np.sum([len(m) for m in mode]) == 0:
                return openSIN_integer_mode(path)
            else:
                return openSIN_old(path)


def openSIN_integer_mode(path):
    """Integer mode file format of e.g. flex03lq-1 correlator (correlator.com)

    This is a file format where the type (AC/CC) of the curve is
    determined using integers in the "Mode=" line, e.g.

         Mode= 2    3    3    2    0    1    1    0

    which means the first correlation is CC23, the second CC32,
    the third CC01, and the fourth CC10. Similarly, 

        Mode= 1    1    2    2    0    4    4    4

    would translate to AC11, AC22, CC04, and AC44.
    """
    with path.open() as fd:
        data = fd.readlines()

    # get mode (curve indices)
    for line in data:
        line = line.strip()
        if line.lower().startswith("mode"):
            mode = line.split("=")[1].strip().split()
            mode = [int(m) for m in mode]
            if len(mode) % 2 != 0:
                msg = "mode must be multiples of two: {}".format(path)
                raise OpenSINError(msg)

    # build up the lists
    corr_func = []
    intensity = []
    section = ""

    # loop through lines
    for line in data:
        line = line.strip().lower()
        if line.startswith("["):
            section = line
            continue
        elif (len(line) == 0 or
              line.count("=")):
            continue

        if section.count("[correlationfunction]"):
            corr_func.append(line.split())
        elif section.count("[intensityhistory]"):
            intensity.append(line.split())

    # corr_func now contains lag time, and correlations according
    # to the mode parameters.
    corr_func = np.array(corr_func, dtype=float)
    intensity = np.array(intensity, dtype=float)

    timefactor = 1000  # because we want ms instead of s
    timedivfac = 1000  # because we want kHz instead of Hz
    intensity[:, 0] *= timefactor
    corr_func[:, 0] *= timefactor
    intensity[:, 1:] /= timedivfac

    # correlator.com correlation is normalized to 1, not to 0
    corr_func[:, 1:] -= 1

    # Now sort the information for pycorrfit
    correlations = []
    traces = []
    curvelist = []

    for ii in range(len(mode)//2):
        modea = mode[2*ii]
        modeb = mode[2*ii+1]

        if modea == modeb:
            # curve type AC
            curvelist.append("AC{}".format(modea))
            # trace
            atrace = np.zeros((intensity.shape[0], 2), dtype=float)
            atrace[:, 0] = intensity[:, 0]
            atrace[:, 1] = intensity[:, modea+1]
            traces.append(atrace)
        else:
            # curve type CC
            curvelist.append("CC{}{}".format(modea, modeb))
            # trace
            modmin = min(modea, modeb)
            modmax = max(modea, modeb)
            tracea = np.zeros((intensity.shape[0], 2), dtype=float)
            tracea[:, 0] = intensity[:, 0]
            tracea[:, 1] = intensity[:, modmin+1]
            traceb = np.zeros((intensity.shape[0], 2), dtype=float)
            traceb[:, 0] = intensity[:, 0]
            traceb[:, 1] = intensity[:, modmax+1]
            traces.append([tracea, traceb])
        # correlation
        corr = np.zeros((corr_func.shape[0], 2), dtype=float)
        corr[:, 0] = corr_func[:, 0]
        corr[:, 1] = corr_func[:, ii+1]
        correlations.append(corr)

    dictionary = {}
    dictionary["Correlation"] = correlations
    dictionary["Trace"] = traces
    dictionary["Type"] = curvelist
    filelist = []
    for _i in curvelist:
        filelist.append(path.name)
    dictionary["Filename"] = filelist
    return dictionary


def openSIN_old(path):
    """Parses the simple sin file format (correlator.com)

    Read data from a .SIN file, usually created by
    the software using correlators from correlator.com.

        FLXA
        Version= 1d

        [Parameters]
        ...
        Mode= Single Auto
        ...

        [CorrelationFunction]
        1.562500e-09	0.000000e+00
        3.125000e-09	0.000000e+00
        4.687500e-09	0.000000e+00
        ...
        1.887435e+01	1.000030e+00
        1.929378e+01	1.000141e+00
        1.971321e+01	9.999908e-01
        2.013264e+01	9.996810e-01
        2.055207e+01	1.000047e+00
        2.097150e+01	9.999675e-01
        2.139093e+01	9.999591e-01
        2.181036e+01	1.000414e+00
        2.222979e+01	1.000129e+00
        2.264922e+01	9.999285e-01
        2.306865e+01	1.000077e+00
        ...
        3.959419e+02	0.000000e+00
        4.026528e+02	0.000000e+00
        4.093637e+02	0.000000e+00
        4.160746e+02	0.000000e+00
        4.227854e+02	0.000000e+00
        4.294963e+02	0.000000e+00

        [RawCorrelationFunction]
        ...

        [IntensityHistory]
        TraceNumber= 458
        0.000000	9.628296e+03	9.670258e+03
        0.262144	1.001358e+04	9.971619e+03
        0.524288	9.540558e+03	9.548188e+03
        0.786432	9.048462e+03	9.010315e+03
        1.048576	8.815766e+03	8.819580e+03
        1.310720	8.827210e+03	8.861542e+03
        1.572864	9.201050e+03	9.185791e+03
        1.835008	9.124756e+03	9.124756e+03
        2.097152	9.059906e+03	9.029389e+03
        ...

    1. We are interested in the "[CorrelationFunction]" section,
    where the first column denotes tau in seconds and the second row the
    correlation signal. Values are separated by a tabulator "\t".
    We do not import anything from the "[Parameters]" section.
    We have to subtract "1" from the correlation function, since it
    is a correlation function that converges to "1" and not to "0".

    2. We are also interested in the "[IntensityHistory]" section.
    If we are only interested in autocorrelation functions: An email
    from Jixiang Zhu - Correlator.com (2012-01-22) said, that
    "For autocorrelation mode, the 2nd and 3 column represent the same
    intensity series with slight delay.  Therefore, they are statistically
    the same but numerically different."
    It is therefore perfectly fine to just use the 2nd column.

    Different acquisition modes:
    Mode            [CorrelationFunction]               [IntensityHistory]
    Single Auto     2 Colums (tau,AC)                   1 significant
    Single Cross    2 Colums (tau,CC)                   2
    Dual Auto       3 Colums (tau,AC1,AC2)              2
    Dual Cross      3 Colums (tau,CC12,CC21)            2
    Quad            5 Colums (tau,AC1,AC2,CC12,CC21)    2

    Returns:
    [0]:
     N arrays with tuples containing two elements:
     1st: tau in ms
     2nd: corresponding correlation signal
    [1]:
     N Intensity traces:
     1st: time in ms
     2nd: Trace in kHz
    [2]: 
     A list with N elements, indicating, how many correlation
     curves we are importing.
    """
    with path.open() as fd:
        Alldata = fd.readlines()
    # Find out where the correlation function and trace are
    for i in np.arange(len(Alldata)):
        if Alldata[i][0:4] == "Mode":

            Mode = Alldata[i].split("=")[1].strip()
        if Alldata[i][0:21] == "[CorrelationFunction]":
            StartC = i+1
        if Alldata[i][0:24] == "[RawCorrelationFunction]":
            EndC = i-2
        if Alldata[i][0:18] == "[IntensityHistory]":
            # plus 2, because theres a line with the trace length
            StartT = i+2
        if Alldata[i][0:11] == "[Histogram]":
            EndT = i-2
    curvelist = []
    correlations = []
    traces = []
    # Get the correlation function
    Truedata = Alldata[StartC:EndC]
    timefactor = 1000  # because we want ms instead of s
    readcorr = csv.reader(Truedata, delimiter='\t')
    # Trace
    # Trace is stored in three columns
    # 1st column: time [s]
    # 2nd column: trace [Hz]
    # 3rd column: trace [Hz] - Single Auto: equivalent to 2nd
    # Get the trace
    Tracedata = Alldata[StartT:EndT]
    # timefactor = 1000 # because we want ms instead of s
    timedivfac = 1000  # because we want kHz instead of Hz
    readtrace = csv.reader(Tracedata, delimiter='\t')
    # Process all Data:
    if Mode == "Single Auto":
        curvelist.append("AC")
        corrdata = []
        for row in readcorr:
            # tau in ms, corr-function minus "1"
            corrdata.append(
                (np.float64(row[0])*timefactor, np.float64(row[1])-1))
        correlations.append(np.array(corrdata))
        trace = []
        for row in readtrace:
            # tau in ms, corr-function minus "1"
            trace.append((np.float64(row[0])*timefactor,
                          np.float64(row[1])/timedivfac))
        traces.append(np.array(trace))
    elif Mode == "Single Cross":
        curvelist.append("CC")
        corrdata = []
        for row in readcorr:
            # tau in ms, corr-function minus "1"
            corrdata.append(
                (np.float64(row[0])*timefactor, np.float64(row[1])-1))
        correlations.append(np.array(corrdata))
        trace1 = []
        trace2 = []
        for row in readtrace:
            # tau in ms, corr-function minus "1"
            trace1.append((np.float64(row[0])*timefactor,
                           np.float64(row[1])/timedivfac))
            trace2.append((np.float64(row[0])*timefactor,
                           np.float64(row[2])/timedivfac))
        traces.append([np.array(trace1), np.array(trace2)])
    elif Mode == "Dual Auto":
        curvelist.append("AC1")
        curvelist.append("AC2")
        corrdata1 = []
        corrdata2 = []
        for row in readcorr:
            # tau in ms, corr-function minus "1"
            corrdata1.append(
                (np.float64(row[0])*timefactor, np.float64(row[1])-1))
            corrdata2.append(
                (np.float64(row[0])*timefactor, np.float64(row[2])-1))
        correlations.append(np.array(corrdata1))
        correlations.append(np.array(corrdata2))
        trace1 = []
        trace2 = []
        for row in readtrace:
            # tau in ms, corr-function minus "1"
            trace1.append((np.float64(row[0])*timefactor,
                           np.float64(row[1])/timedivfac))
            trace2.append((np.float64(row[0])*timefactor,
                           np.float64(row[2])/timedivfac))
        traces.append(np.array(trace1))
        traces.append(np.array(trace2))
    elif Mode == "Dual Cross":
        curvelist.append("CC12")
        curvelist.append("CC21")
        corrdata1 = []
        corrdata2 = []
        for row in readcorr:
            # tau in ms, corr-function minus "1"
            corrdata1.append(
                (np.float64(row[0])*timefactor, np.float64(row[1])-1))
            corrdata2.append(
                (np.float64(row[0])*timefactor, np.float64(row[2])-1))
        correlations.append(np.array(corrdata1))
        correlations.append(np.array(corrdata2))
        trace1 = []
        trace2 = []
        for row in readtrace:
            # tau in ms, corr-function minus "1"
            trace1.append((np.float64(row[0])*timefactor,
                           np.float64(row[1])/timedivfac))
            trace2.append((np.float64(row[0])*timefactor,
                           np.float64(row[2])/timedivfac))
        traces.append([np.array(trace1), np.array(trace2)])
        traces.append([np.array(trace1), np.array(trace2)])
    elif Mode == "Quad":
        curvelist.append("AC1")
        curvelist.append("AC2")
        curvelist.append("CC12")
        curvelist.append("CC21")
        corrdata1 = []
        corrdata2 = []
        corrdata12 = []
        corrdata21 = []
        for row in readcorr:
            # tau in ms, corr-function minus "1"
            corrdata1.append(
                (np.float64(row[0])*timefactor, np.float64(row[1])-1))
            corrdata2.append(
                (np.float64(row[0])*timefactor, np.float64(row[2])-1))
            corrdata12.append(
                (np.float64(row[0])*timefactor, np.float64(row[3])-1))
            corrdata21.append(
                (np.float64(row[0])*timefactor, np.float64(row[4])-1))
        correlations.append(np.array(corrdata1))
        correlations.append(np.array(corrdata2))
        correlations.append(np.array(corrdata12))
        correlations.append(np.array(corrdata21))
        trace1 = []
        trace2 = []
        for row in readtrace:
            # tau in ms, corr-function minus "1"
            trace1.append((np.float64(row[0])*timefactor,
                           np.float64(row[1])/timedivfac))
            trace2.append((np.float64(row[0])*timefactor,
                           np.float64(row[2])/timedivfac))
        traces.append(np.array(trace1))
        traces.append(np.array(trace2))
        traces.append([np.array(trace1), np.array(trace2)])
        traces.append([np.array(trace1), np.array(trace2)])
    else:
        raise NotImplemented(
            "'Mode' type '{}' in {} not supported by this method!".
            Mode, format(path))

    dictionary = {}
    dictionary["Correlation"] = correlations
    dictionary["Trace"] = traces
    dictionary["Type"] = curvelist
    filelist = []
    for i in curvelist:
        filelist.append(path.name)
    dictionary["Filename"] = filelist

    return dictionary



"""PyCorrFit data set: Classes for FCS data evaluation"""
import copy
import warnings

import lmfit
import numpy as np
import scipy.interpolate as spintp


class StuckParameterWarning(UserWarning):
    pass


class Constraint(object):
    """ Class to translate fit constraints to lmfit syntax.
    """

    def __init__(self, constraint, fit_bool, fit_bounds, fit_values):
        """
        Parameters
        ----------
        constraint : list of strings and ints
            The abstract constraint (e.g. [1, 0, "<", "2.3"]) as
            used in the model definitions.
        fit_bool : list of boolean
            A list of bools indicating which parameters are varied
            during fitting.
        fit_bounds : list of lists of two floats
            The parameter boundaries for fitting.
        fit_values : list of floats
            The initial fitting values.

        Notes
        -----
        - the first item in constraints must be an integer indexing a parameter
        - the second/third item must be an integer as well or an operator (">", "<")
        - if the fourth item is omitted, it is assumed to be "0"
        - the first integer must be larger than the second integer
        """
        if len(constraint) == 3:
            constraint.append("0")

        self.constraint = constraint
        self.fit_bool = fit_bool
        self.fit_bounds = fit_bounds
        self.fit_values = fit_values

    @property
    def parameters(self):
        """
        Returns list of dict for each parameter.
        """
        parms = [it for it in self.constraint if isinstance(it, int)]
        id2 = self.constraint.index(parms[1])

        p1 = {"id": parms[0],
              "bool": self.fit_bool[parms[0]],
              "sign": +1,
              "value": self.fit_values[parms[0]]
              }

        p2 = {"id": parms[1],
              "bool": self.fit_bool[parms[1]],
              "sign": (+1 if id2 == 2 else -1),
              "value": self.fit_values[parms[1]]
              }

        return p1, p2

    @property
    def operator(self):
        strval = [it for it in self.constraint if not isinstance(it, int)]
        return strval[0]

    @property
    def offset(self):
        return float(self.constraint[-1])

    def update_fit_bounds(self):
        """
        Update the bounds with the given constraint. This only applies
        if one of the parameters is not varied during fitting.

        Notes
        -----
        The fitting boundaries are updated in-place (`fit_bounds` variable
        from `__init__`).
        """
        p1, p2 = self.parameters
        op = self.operator
        os = self.offset

        if op not in ["<", ">"]:
            raise ValueError("Unsupported operator: {}".format(op))

        if p1["bool"] and p2["bool"]:
            # do nothing, this case is handled in `get_lmfit_parameter_kwargs`
            pass
        elif p1["bool"]:
            # only parameter 1 is varied
            if op == "<":
                # [3, "<", 1, "0"] -> p1 < p2
                # [3, 1, "<", "0"] -> p1 < -p2
                # [3, 1, "<", "1.2] -> p1 < -p2 + 1.2
                # [3, "<", 1, "1.2] -> p1 < p2 + 1.2
                bnd = [-np.inf, p2["sign"]*p2["value"] + os]
            else:
                # [3, ">", 1, "0"] -> p1 > p2
                # [3, 1, ">", "0"] -> p1 > -p2
                # [3, 1, ">", "1.2] -> p1 > -p2 + 1.2
                # [3, ">", 1, "1.2] -> p1 > p2 + 1.2
                bnd = [p2["sign"]*p2["value"] + os, np.inf]
            bound = [max(self.fit_bounds[p1["id"]][0], bnd[0]),
                     min(self.fit_bounds[p1["id"]][1], bnd[1]),
                     ]
            self.fit_bounds[p1["id"]] = bound
        elif p2["bool"]:
            # only parameter2 is varied
            if op == "<":
                # [3, "<", 1, "0"] -> p2 > p1
                # [3, 1, "<", "0"] -> (-)p2 > p1 -> p2 < -p1
                # [3, 1, "<", "1.2] -> p2 < -p1 + 1.2 = -(p1-1.2)
                # [3, "<", 1, "1.2] -> p2 > p1 - 1.2 = +(p1-1.2)
                if p2["sign"] == -1:
                    bnd = [-np.inf, -(p1["value"] - os)]
                else:
                    bnd = [(p1["value"] - os), np.inf]
            else:
                # [3, ">", 1, "0"] -> p2 < p1
                # [3, 1, ">", "0"] -> p2 > -p1
                # [3, 1, ">", "1.2] -> p2 > -(p1 - 1.2)
                # [3, ">", 1, "1.2] -> p2 < p1 - 1.2
                if p2["sign"] == -1:
                    bnd = [-(p1["value"] - os), np.inf]
                else:
                    bnd = [-np.inf, p1["value"] - os]
            bound = [max(self.fit_bounds[p2["id"]][0], bnd[0]),
                     min(self.fit_bounds[p2["id"]][1], bnd[1]),
                     ]
            self.fit_bounds[p2["id"]] = bound
        else:
            # neither is varied.
            # Do nothing.
            pass
        return self.fit_bounds

    def get_lmfit_parameter_kwargs(self):
        """
        Using the given constraint, update the list of lmfit
        parameters.
        """
        p1, p2 = self.parameters
        op = self.operator
        ofs = self.offset

        if op not in ["<", ">"]:
            raise ValueError("Unsupported operator: {}".format(op))

        if p1["bool"] and p2["bool"]:
            if op == "<":
                # p1 < (-)p2 + 1.2
                # -> p1 = (-)p2 - d12 + 1.2
                # -> d12 = (-)p2 - p1 + 1.2
                # -> d12 > 0
                deltaname = "delta_{}_{}".format(p1["id"], p2["id"])
                kwdelt = {}
                kwdelt["name"] = deltaname
                kwdelt["value"] = p2["bool"] * \
                    self.fit_values[p2["id"]] - self.fit_values[p1["id"]]
                kwdelt["vary"] = True
                kwdelt["min"] = 0  # note: enforces "<=" (not "<")
                kwdelt["max"] = np.inf

                kwp1 = {}
                kwp1["name"] = "parm{:04d}".format(p1["id"])
                # this condition deals with negative numbers
                kwp1["expr"] = "{MIN} if {COMP} < {MIN} else {MAX} if {COMP} > {MAX} else {COMP}".format(
                    COMP="{}*parm{:04d}-{}+{:.14f}".format(
                        p2["sign"], p2["id"], deltaname, ofs),
                    MIN=self.fit_bounds[p1["id"]][0],
                    MAX=self.fit_bounds[p1["id"]][1])
                kwargs = [kwdelt, kwp1]
            elif op == ">":
                # p1 > (-)p2 + 1.2
                # -> p1 = (-)p2 + d12 + 1.2
                # -> d12 = p1 - (-)p2  - 1.2
                # -> d12 > 0
                deltaname = "delta_{}_{}".format(p1["id"], p2["id"])
                kwdelt = {}
                kwdelt["name"] = deltaname
                kwdelt["value"] = self.fit_values[p1["id"]] - \
                    p2["bool"]*self.fit_values[p2["id"]]
                kwdelt["vary"] = True
                kwdelt["min"] = 0  # note: enforces ">=" (not ">")
                # self.fit_bounds[p1["id"]][1] + max(-p2["sign"]*self.fit_bounds[p2["id"]]) - ofs
                kwdelt["max"] = np.inf

                kwp1 = {}
                kwp1["name"] = "parm{:04d}".format(p1["id"])
                # this condition deals with negative numbers
                kwp1["expr"] = "{MIN} if {COMP} < {MIN} else {MAX} if {COMP} > {MAX} else {COMP}".format(
                    COMP="{}*parm{:04d}+{}+{:.14f}".format(
                        p2["sign"], p2["id"], deltaname, ofs),
                    MIN=self.fit_bounds[p1["id"]][0],
                    MAX=self.fit_bounds[p1["id"]][1])

                kwargs = [kwdelt, kwp1]

        else:
            kwargs = None

        return kwargs


class Fit(object):
    """ Used for fitting FCS data to models.
    """

    def __init__(self, correlations=[], global_fit=False,
                 global_fit_variables=[],
                 uselatex=False, verbose=0):
        """ Using an FCS model, fit the data of shape (N,2).


        Parameters
        ----------
        correlations: list of instances or instance of `pycorrfit.Correlation`
            Correlations to fit.
        global fit : bool
            Perform global fit. The default behavior is
            to fit all parameters that are selected for
            fitting in each correlation. Parameters with
            the same name in different models are treated
            as one global parameter.
        global_fit_variables: list of list of strings
            Each item contains a list of strings that are names
            of parameters which will be treated as a common
            parameter. This breaks the defaul behavior.
            NOT IMPLEMENTED YET!
        verbose: int
            Increase verbosity by incrementing this number.
        uselatex: bool
            If verbose > 0, plotting will be performed with LaTeX.
        """
        if len(global_fit_variables) != 0:
            raise NotImplementedError("`global_fit_variables` not available!")

        if not isinstance(correlations, list):
            correlations = [correlations]

        self.correlations = correlations
        self.global_fit_variables = global_fit_variables
        self.verbose = verbose
        self.uselatex = uselatex
        self.is_weighted_fit = False

        if not global_fit:
            # Fit each correlation separately
            for corr in self.correlations:
                # Set fitting options
                self.fit_algorithm = corr.fit_algorithm
                # Get the data required for fitting
                self.x = corr.correlation_fit[:, 0]
                self.y = corr.correlation_fit[:, 1]
                # fit_bool: True for variable
                self.fit_bool = corr.fit_parameters_variable.copy()
                self.fit_parm = corr.fit_parameters.copy()
                self.fit_bound = copy.copy(corr.fit_parameters_range)
                self.is_weighted_fit = corr.is_weighted_fit
                self.fit_weights = Fit.compute_weights(corr,
                                                       verbose=verbose,
                                                       uselatex=uselatex)
                self.fit_parm_names = corr.fit_model.parameters[0]
                self.func = corr.fit_model.function
                self.check_parms = corr.check_parms
                self.constraints = corr.fit_model.constraints
                # Directly perform the fit and set the "fit" attribute
                self.minimize()
                # Run a second time:
                self.minimize()
                # update correlation model parameters
                corr.fit_parameters = self.fit_parm
                # save fit data in correlation class
                corr.fit_results = self.get_fit_results(corr)
        else:
            # TODO:
            # - allow detaching of parameters,
            #   i.e. fitting "n" separately for two models
            # Initiate all arrays
            self.fit_algorithm = self.correlations[0].fit_algorithm
            xtemp = []      # x
            ytemp = []      # y
            weights = []    # weights
            ids = [0]           # ids in big fitting array
            cmodels = []    # correlation model info
            initpar = []    # initial parameters
            varin = []      # names of variable fitting parameters
            variv = []      # values of variable fitting parameters
            varmap = []     # list of indices of fitted parameters
            varbound = []   # list of fitting boundaries
            self.is_weighted_fit = None
            for corr in self.correlations:
                xtemp.append(corr.correlation_fit[:, 0])
                ytemp.append(corr.correlation_fit[:, 1])
                weights.append(Fit.compute_weights(corr))
                ids.append(len(xtemp[-1])+ids[-1])
                cmodels.append(corr.fit_model)
                initpar.append(corr.fit_parameters)
                # Create list of variable parameters
                varthis = []
                for ipm, par in enumerate(corr.fit_model.parameters[0]):
                    if corr.fit_parameters_variable[ipm]:
                        varthis.append(ipm)
                        varin.append(par)
                        variv.append(corr.fit_parameters[ipm])
                        varbound.append(corr.fit_parameters_range[ipm])
                varmap.append(varthis)

            # These are the variable fitting parameters
            __, varidx = np.unique(varin, return_index=True)
            varidx.sort()
            varin = np.array(varin)[varidx]
            variv = np.array(variv)[varidx]

            self.x = np.concatenate(xtemp)
            self.y = np.concatenate(ytemp)
            self.fit_bool = np.ones(len(variv), dtype=bool)
            self.fit_parm = variv
            self.fit_weights = np.concatenate(weights)
            self.fit_parm_names = varin
            self.fit_bound = varbound
            self.constraints = []
            warnings.warn(
                "Constraints are not supported yet for global fitting.")

            def parameters_global_to_local(parameters, iicorr, varin=varin,
                                           initpar=initpar,
                                           correlations=correlations):
                """
                With global `parameters` and an id `iicorr` pointing at
                the correlation in `self.correlations`, return the
                updated parameters of the corresponding model.
                """
                fit_parm = initpar[iicorr].copy()
                corr = correlations[iicorr]
                mod = corr.fit_model
                for kk, pn in enumerate(mod.parameters[0]):
                    if pn in varin:
                        # edit that parameter
                        fit_parm[kk] = parameters[np.where(
                            np.array(varin) == pn)[0]][0]
                return fit_parm

            def parameters_local_to_global(parameters, iicorr, fit_parm,
                                           varin=varin,
                                           correlations=correlations):
                """
                inverse of parameters_global_to_local
                """
                corr = correlations[iicorr]
                mod = corr.fit_model
                for kk, pn in enumerate(mod.parameters[0]):
                    if pn in varin:
                        # edit that parameter
                        parameters[np.where(np.array(varin) == pn)[
                            0]] = fit_parm[kk]
                return parameters

            # Create function for fitting using ids
            def global_func(parameters, tau,
                            glob2loc=parameters_global_to_local):
                out = []
                # ids start at 0
                for ii, mod in enumerate(cmodels):
                    # Update parameters
                    fit_parm = glob2loc(parameters, ii)
                    # return function
                    out.append(mod.function(fit_parm, tau[ids[ii]:ids[ii+1]]))
                return np.concatenate(out)

            self.func = global_func

            # Create function for checking
            def global_check_parms(parameters,
                                   glob2loc=parameters_global_to_local,
                                   loc2glob=parameters_local_to_global):

                for ii, corr in enumerate(self.correlations):
                    # create new initpar
                    fit_parm = glob2loc(parameters, ii)
                    fit_parm = corr.check_parms(fit_parm)
                    # update parameters
                    parameters = loc2glob(parameters, ii, fit_parm)

                return parameters

            self.check_parms = global_check_parms

            # Directly perform the fit and set the "fit" attribute
            self.minimize()
            # Update correlations
            for ii, corr in enumerate(self.correlations):
                # write new model parameters
                corr.fit_parameters = parameters_global_to_local(self.fit_parm,
                                                                 ii)
                # save fit data in correlation class
                corr.fit_results = self.get_fit_results(corr)

    def get_fit_results(self, correlation):
        """
        Return a dictionary with all information about the performed fit.

        This function must be called immediately after `self.minimize`.
        """
        c = correlation
        d = {
            "chi2": self.chi_squared,
            "chi2 type": self.chi_squared_type,
            "weighted fit": c.is_weighted_fit,
            "fit algorithm": c.fit_algorithm,
            "fit result": 1*c.fit_parameters,
            "fit parameters": 1*np.where(c.fit_parameters_variable)[0],
            "fit weights": 1*self.compute_weights(c)
        }

        if c.is_weighted_fit:
            d["weighted fit type"] = c.fit_weight_type
            if isinstance(c.fit_weight_data, (int, float)):
                d["weighted fit bins"] = c.fit_weight_data

        if d["fit algorithm"] == "Lev-Mar" and self.parmoptim_error is not None:
            d["fit error estimation"] = self.parmoptim_error

        return d

    @property
    def chi_squared(self):
        """ Calculate displayed Chi²

            Calculate reduced Chi² for the current class.
        """
        # Calculate degrees of freedom
        dof = len(self.x) - np.sum(self.fit_bool) - 1
        # This is exactly what is minimized by the scalar minimizers

        if self.chi_squared_type == "reduced expected sum of squares":
            fitted = self.func(self.fit_parm, self.x)
            chi2 = np.sum((self.y-fitted)**2/np.abs(fitted)) / dof
        elif self.chi_squared_type == "reduced weighted sum of squares":
            fitted = self.func(self.fit_parm, self.x)
            variance = self.fit_weights**2
            chi2 = np.sum((self.y-fitted)**2/variance) / dof
        elif self.chi_squared_type == "reduced global sum of squares":
            fitted = self.func(self.fit_parm, self.x)
            variance = self.fit_weights**2
            chi2 = np.sum((self.y-fitted)**2/variance) / dof
        else:
            chi2 = self.fit_function_scalar(
                self.fit_parm, self.x, self.y, self.fit_weights)

        return chi2

    @property
    def chi_squared_type(self):
        """ The type of Chi² that currently applies.

        Returns
        -------
        "reduced" - if variance of data was used for fitting
        "reduced Pearson" - if variance of data is not available
        """
        if self.is_weighted_fit is None:
            # global fitting
            return "reduced global sum of squares"
        elif self.is_weighted_fit == True:
            return "reduced weighted sum of squares"
        elif self.is_weighted_fit == False:
            return "reduced expected sum of squares"
        else:
            raise ValueError("Unknown weight type!")

    @staticmethod
    def compute_weights(correlation, verbose=0, uselatex=False):
        """ computes and returns weights of the same length as
        `correlation.correlation_fit`

        `correlation` is an instance of Correlation
        """
        corr = correlation
        model = corr.fit_model
        model_parms = corr.fit_parameters
        ival = corr.fit_ival
        weight_data = corr.fit_weight_data
        weight_type = corr.fit_weight_type
        # parameters = corr.fit_parameters
        # parameters_range = corr.fit_parameters_range
        # parameters_variable = corr.fit_parameters_variable

        cdat = corr.correlation
        if cdat is None:
            raise ValueError("Cannot compute weights; No correlation given!")
        cdatfit = corr.correlation_fit
        x_full = cdat[:, 0]
        y_full = cdat[:, 1]
        x_fit = cdatfit[:, 0]
        # y_fit = cdatfit[:,1]

        dataweights = np.ones_like(x_fit)

        try:
            weight_spread = int(weight_data)
        except:
            if verbose > 1:
                warnings.warn(
                    "Could not get weight spread for spline. Setting it to 3.")
            weight_spread = 3

        if weight_type[:6] == "spline":
            # Number of knots to use for spline
            try:
                knotnumber = int(weight_type[6:])
            except:
                if verbose > 1:
                    print("Could not get knot number. Setting it to 5.")
                knotnumber = 5

            # Compute borders for spline fit.
            if ival[0] < weight_spread:
                # optimal case
                pmin = ival[0]
            else:
                # non-optimal case
                # we need to cut pmin
                pmin = weight_spread
            if x_full.shape[0] - ival[1] < weight_spread:
                # optimal case
                pmax = x_full.shape[0] - ival[1]
            else:
                # non-optimal case
                # we need to cut pmax
                pmax = weight_spread

            x = x_full[ival[0]-pmin:ival[1]+pmax]
            y = y_full[ival[0]-pmin:ival[1]+pmax]
            # we are fitting knots on a base 10 logarithmic scale.
            xs = np.log10(x)
            knots = np.linspace(xs[1], xs[-1], knotnumber+2)[1:-1]
            try:
                tck = spintp.splrep(xs, y, s=0, k=3, t=knots, task=-1)
                ys = spintp.splev(xs, tck, der=0)
            except:
                if verbose > 0:
                    raise ValueError("Could not find spline fit with " +
                                     "{} knots.".format(knotnumber))
                return
            if verbose > 0:
                # If plotting module is available:
                # name = "spline fit: "+str(knotnumber)+" knots"
                # plotting.savePlotSingle(name, 1*x, 1*y, 1*ys,
                #                         dirname=".",
                #                         uselatex=uselatex)
                # use matplotlib.pylab
                try:
                    from matplotlib import pylab as plt
                    plt.xscale("log")
                    plt.plot(x, ys, x, y)
                    plt.show()
                except ImportError:
                    # Tell the user to install matplotlib
                    print("Couldn't import pylab! - not Plotting")

            # Calculation of variance
            # In some cases, the actual cropping interval from ival[0]
            # to ival[1] is chosen, such that the dataweights must be
            # calculated from unknown datapoints.
            # (e.g. points+endcrop > len(correlation)
            # We deal with this by multiplying dataweights with a factor
            # corresponding to the missed points.
            for i in range(x_fit.shape[0]):
                # Define start and end positions of the sections from
                # where we wish to calculate the dataweights.
                # Offset at beginning:
                if i + ival[0] < weight_spread:
                    # The offset that occurs
                    offsetstart = weight_spread - i - ival[0]
                    offsetcrop = 0
                elif ival[0] > weight_spread:
                    offsetstart = 0
                    offsetcrop = ival[0] - weight_spread
                else:
                    offsetstart = 0
                    offsetcrop = 0
                # i: counter on correlation array
                # start: counter on y array
                start = i - weight_spread + offsetstart + ival[0] - offsetcrop
                end = start + 2*weight_spread + 1 - offsetstart
                dataweights[i] = (y[start:end] - ys[start:end]).std()
                # The standard deviation at the end and the start of the
                # array are multiplied by a factor corresponding to the
                # number of bins that were not used for calculation of the
                # standard deviation.
                if offsetstart != 0:
                    reference = 2*weight_spread + 1
                    dividor = reference - offsetstart
                    dataweights[i] *= reference/dividor
                # Do not substitute len(y[start:end]) with end-start!
                # It is not the same!
                backset = 2*weight_spread + 1 - len(y[start:end]) - offsetstart
                if backset != 0:
                    reference = 2*weight_spread + 1
                    dividor = reference - backset
                    dataweights[i] *= reference/dividor
        elif weight_type == "model function":
            # Number of neighboring (left and right) points to include
            if ival[0] < weight_spread:
                pmin = ival[0]
            else:
                pmin = weight_spread
            if x_full.shape[0] - ival[1] < weight_spread:
                pmax = x_full.shape[0] - ival[1]
            else:
                pmax = weight_spread
            x = x_full[ival[0]-pmin:ival[1]+pmax]
            y = y_full[ival[0]-pmin:ival[1]+pmax]
            # Calculated dataweights
            for i in np.arange(x_fit.shape[0]):
                # Define start and end positions of the sections from
                # where we wish to calculate the dataweights.
                # Offset at beginning:
                if i + ival[0] < weight_spread:
                    # The offset that occurs
                    offsetstart = weight_spread - i - ival[0]
                    offsetcrop = 0
                elif ival[0] > weight_spread:
                    offsetstart = 0
                    offsetcrop = ival[0] - weight_spread
                else:
                    offsetstart = 0
                    offsetcrop = 0
                # i: counter on correlation array
                # start: counter on correlation array
                start = i - weight_spread + offsetstart + ival[0] - offsetcrop
                end = start + 2*weight_spread + 1 - offsetstart
                # start = ival[0] - weight_spread + i
                # end = ival[0] + weight_spread + i + 1
                diff = y - model(model_parms, x)
                dataweights[i] = diff[start:end].std()
                # The standard deviation at the end and the start of the
                # array are multiplied by a factor corresponding to the
                # number of bins that were not used for calculation of the
                # standard deviation.
                if offsetstart != 0:
                    reference = 2*weight_spread + 1
                    dividor = reference - offsetstart
                    dataweights[i] *= reference/dividor
                # Do not substitute len(diff[start:end]) with end-start!
                # It is not the same!
                backset = 2*weight_spread + 1 - \
                    len(diff[start:end]) - offsetstart
                if backset != 0:
                    reference = 2*weight_spread + 1
                    dividor = reference - backset
                    dataweights[i] *= reference/dividor
        elif weight_type == "none":
            pass
        else:
            # This means that the user knows the dataweights and already
            # gave it to us.
            weights = weight_data
            if weights is None:
                msg = "User defined weights not given: {}".format(weight_type)
                raise ValueError(msg)

            # Check if these other weights have length of the cropped
            # or the full array.
            if weights.shape[0] == x_fit.shape[0]:
                dataweights = weights
            elif weights.shape[0] == x_full.shape[0]:
                dataweights = weights[ival[0]:ival[1]]
            else:
                msg = "`weights` must have length of full or cropped array."
                raise ValueError(msg)

        return dataweights

    def fit_function(self, params, x, y, weights=1):
        """
        objective function that returns the residual (difference between
        model and data) to be minimized in a least squares sense.
        """
        parms = Fit.lmfitparm2array(params)
        tominimize = (self.func(parms, x) - y)
        # Check dataweights for zeros and don't use these
        # values for the least squares method.
        with np.errstate(divide='ignore'):
            tominimize = np.where(weights != 0,
                                  tominimize/weights, 0)
        # There might be NaN values because of zero weights:
        # tominimize = tominimize[~np.isinf(tominimize)]
        return tominimize

    def fit_function_scalar(self, parms, x, y, weights=1):
        """
        Wrapper of `fit_function`.
        Returns the sum of squares of the input data.
        """
        e = self.fit_function(parms, x, y, weights)
        return np.sum(e*e)

    def get_lmfitparm(self):
        """
        Generates an lmfit parameter class from the present data set.

        The following parameters are used:
        self.x : 1d ndarray length N
        self.y : 1d ndarray length N
        self.fit_weights : 1d ndarray length N

        self.fit_bool : 1d ndarray length P, bool
        self.fit_parm : 1d ndarray length P, float
        """
        params = lmfit.Parameters()

        # First, add all fixed parameters
        for pp in range(len(self.fit_parm)):
            if not self.fit_bool[pp]:
                if self.fit_bound[pp][0] == self.fit_bound[pp][1]:
                    self.fit_bound[pp] = [-np.inf, np.inf]
                params.add(lmfit.Parameter(name="parm{:04d}".format(pp),
                                           value=self.fit_parm[pp],
                                           vary=self.fit_bool[pp],
                                           min=self.fit_bound[pp][0],
                                           max=self.fit_bound[pp][1],
                                           )
                           )

        # Second, summarize the constraints in a dictionary, where
        # keys are the parameter indexes of varied parameters.
        # The dictionary cstrnew only allows integer keys that are
        # representing parameter indices. The fact that we are effectively
        # reducing the number of valid constraints to one per parameter
        # is a design problem that cannot be resolved here. The constraints
        # must be defined in such a way, that a parameter with a larger
        # index number is dependent on only one parameter with a lower
        # index number, e.g. parm1>parm0, parm3<parm1, etc..
        #
        # Constraints have the three-element form:
        #   [1, "<", 0] -> parm1 < parm0
        #   [3, ">", 1] -> parm3 > parm1
        # or the four-element form:
        #   [1, "<", 0, "2.3"]] -> parm1 < parm0 + 2.3
        #   [1, 0, "<", "2.3"]] -> parm1 + parm0 < 2.3
        for pp in range(len(self.fit_parm)):
            if self.fit_bool[pp]:
                inconstr = len([cc for cc in self.constraints if pp in cc])
                kwarglist = []

                if inconstr:
                    for cc in self.constraints:
                        con = Constraint(constraint=cc,
                                         fit_bool=self.fit_bool,
                                         fit_bounds=self.fit_bound,
                                         fit_values=self.fit_parm)
                        self.fit_bound = con.update_fit_bounds()
                        if con.parameters[0]["id"] == pp:
                            kws = con.get_lmfit_parameter_kwargs()
                            if kws is not None:
                                kwarglist += kws
                if len(kwarglist) == 0:
                    # normal parameter
                    kwarglist += [{"name": "parm{:04d}".format(pp),
                                   "value": self.fit_parm[pp],
                                   "vary": True,
                                   "min": self.fit_bound[pp][0],
                                   "max": self.fit_bound[pp][1],
                                   }]
                for kw in kwarglist:
                    params.add(lmfit.Parameter(**kw))

        return params

    @staticmethod
    def lmfitparm2array(parms, parmid="parm", attribute="value"):
        """
        Convert lmfit parameters to a numpy array.
        Parameters are identified by name `parmid` which should
        be at the beginning of a parameters.

        This method is necessary to separate artificial constraint parameters
        from the actual parameters.

        Parameters
        ----------
        parms : lmfit.parameter.Parameters or ndarray
            The input parameters.
        parmid : str
            The identifier for parameters. By default this is
            "parm", i.e. parameters are named like this:
            "parm0001", "parm0002", etc.
        attribute : str
            The attribute to return, e.g.
            - "value" : return the current value of the parameter
            - "vary" : return if the parameter is varied during fitting

        Returns:
        parr : ndarray
            If the input is an ndarray, the input will be returned.
        """
        if isinstance(parms, lmfit.parameter.Parameters):
            items = parms.items()
            parr = []
            for p in sorted(items, key=lambda x: x[0]):
                if p[0].startswith(parmid):
                    parr.append(getattr(p[1], attribute))
        else:
            parr = parms

        return np.array(parr)

    def minimize(self):
        """ This will run the minimization process

        """
        if np.sum(self.fit_bool) == 0:
            raise ValueError("No parameter selected for fitting!")

        # get all parameters for minimization
        params = self.get_lmfitparm()

        # Get algorithm
        method = Algorithms[self.fit_algorithm][0]
        methodkwargs = Algorithms[self.fit_algorithm][2]

        # Begin fitting
        # Fit a several times and stop earlier if the residuals
        # are small enough (heuristic approach).
        nfits = 5
        diff = np.inf
        parmsinit = Fit.lmfitparm2array(params)
        for ii in range(nfits):
            res0 = self.fit_function(params, self.x, self.y)
            result = lmfit.minimize(fcn=self.fit_function,
                                    params=params,
                                    method=method,
                                    kws={"x": self.x,
                                         "y": self.y,
                                         "weights": self.fit_weights},
                                    **methodkwargs
                                    )
            params = result.params
            res1 = self.fit_function(params, self.x, self.y)
            diff = np.average(np.abs(res0-res1))

            if hasattr(result, "ier") and not result.errorbars and ii+1 < nfits:
                # This case applies to the Levenberg-Marquardt algorithm
                multby = .5
                # Try to vary stuck fitting parameters
                # the result from the previous fit
                parmsres = Fit.lmfitparm2array(params)
                # the parameters that are varied during fitting
                parmsbool = Fit.lmfitparm2array(params, attribute="vary")
                # The parameters that are stuck
                parmstuck = parmsbool * (parmsinit == parmsres)
                parmsres[parmstuck] *= multby
                # write changes
                self.fit_parm = parmsres
                params = self.get_lmfitparm()
                warnings.warn(u"PyCorrFit detected problems in fitting, " +
                              u"detected a stuck parameter, multiplied " +
                              u"it by {}, and fitted again. ".format(multby) +
                              u"The stuck parameters are: {}".format(
                                  np.array(self.fit_parm_names)[parmstuck]),
                              StuckParameterWarning)
            elif diff < 1e-8:
                # Experience tells us this is good enough.
                break

        # Now write the optimal parameters to our values:
        self.fit_parm = Fit.lmfitparm2array(params)
        # Only allow physically correct parameters
        self.fit_parm = self.check_parms(self.fit_parm)
        # Compute error estimates for fit (Only "Lev-Mar")
        if (self.fit_algorithm == "Lev-Mar"
            and result.success
            # If the covariance matrix cannot be computed in
            # - lmfit <= 0.9.10: result.covar is None
            # - lmfit >= 0.9.11: result.covar is not set
            and hasattr(result, "covar")
                and result.covar is not None):
            # This is the standard way to minimize the data. Therefore,
            # we are a little bit more verbose.
            covar = result.covar
            try:
                self.parmoptim_error = np.diag(covar)
            except:
                warnings.warn("PyCorrFit Warning: Error estimate not " +
                              "possible, because we could not " +
                              "calculate covariance matrix. Please " +
                              "try reducing the number of fitting " +
                              "parameters.")
                self.parmoptim_error = None
        else:
            self.parmoptim_error = None


def GetAlgorithmStringList():
    """
        Get supported fitting algorithms as strings.
        Returns two lists (that are key-sorted) for key and string.
    """
    A = Algorithms
    out1 = []
    out2 = []
    a = list(A.keys())
    a.sort()
    for key in a:
        out1.append(key)
        out2.append(A[key][1])
    return out1, out2


# As of version 0.8.3, we support several minimization methods for
# fitting data to experimental curves.
# These functions must be callable like scipy.optimize.leastsq. e.g.
# res = spopt.leastsq(self.fit_function, self.fitparms[:],
#                     args=(self.x), full_output=1)
Algorithms = dict()

# the original one is the least squares fit "leastsq"
Algorithms["Lev-Mar"] = ["leastsq",
                         "Levenberg-Marquardt",
                         {"ftol": 1.49012e-08,
                          "xtol": 1.49012e-08,
                          }
                         ]

# simplex
Algorithms["Nelder-Mead"] = ["nelder",
                             "Nelder-Mead (downhill simplex)",
                             {}
                             ]

# quasi-Newton method of Broyden, Fletcher, Goldfarb, and Shanno
Algorithms["BFGS"] = ["lbfgsb",
                      "BFGS (quasi-Newton)",
                      {}
                      ]

# modified Powell-method
Algorithms["Powell"] = ["powell",
                        "modified Powell (conjugate direction)",
                        {}
                        ]

# nonliner conjugate gradient method by Polak and Ribiere
Algorithms["SLSQP"] = ["slsqp",
                       "Sequential Linear Squares Programming",
                       {}
                       ]



"""Module misc: Non-science related code.
"""
import codecs

import numpy as np
import wx

# The icon file was created with
# img2py -i -n Main PyCorrFit_icon.png icon.py
from . import icon                         # Contains the program icon


def parseString2Pagenum(parent, string, nodialog=False):
    """ Parse a string with a list of pagenumbers to an integer list with
        page numbers.
        e.g. "1-3,5,7" --> [1,2,3,5,7]
        parent is important
    """
    listFull = string.split(",")
    PageNumbers = list()
    try:
        for item in listFull:
            pagerange = item.split("-")
            start = pagerange[0].strip()
            start = int("".join(filter(type(start).isdigit, start)))
            end = pagerange[-1].strip()
            end = int("".join(filter(type(end).isdigit, end)))
            for i in np.arange(end-start+1)+start:
                PageNumbers.append(i)
        PageNumbers.sort()
        return PageNumbers
    except:
        if nodialog is False:
            errstring = "Invalid syntax in page selection: "+string +\
                        ". Please use a comma separated list with" +\
                        " optional dashes, e.g. '1-3,6,8'."
            try:
                wx.MessageDialog(parent, errstring, "Error",
                                 style=wx.ICON_ERROR | wx.OK | wx.STAY_ON_TOP)
            except:
                raise ValueError(errstring)
        else:
            raise ValueError(errstring)
        return None


def parsePagenum2String(pagenumlist):
    """ Make a string with dashes and commas from a list of pagenumbers.
        e.g. [1,2,3,5,7] --> "1-3,5,7"
    """
    if len(pagenumlist) == 0:
        return ""
    # Make sure we have integers
    newlist = list()
    for num in pagenumlist:
        newlist.append(int(num))
    newlist.sort()
    # begin string
    string = str(newlist[0])
    # iteration through list:
    dash = False
    for i in np.arange(len(newlist)-1)+1:
        if dash == True:
            if newlist[i]-1 == newlist[i-1]:
                pass
            else:
                string += "-"+str(newlist[i-1])+", "+str(newlist[i])
                dash = False
        else:
            if newlist[i]-1 == newlist[i-1]:
                if newlist[i]-2 == newlist[i-2]:
                    dash = True
                elif len(newlist) != i+1 and newlist[i]+1 == newlist[i+1]:
                    dash = True
                else:
                    string += ", "+str(newlist[i])
                    dash = False
            else:
                dash = False
                string += ", "+str(newlist[i])
        # Put final number
        if newlist[i] == newlist[-1]:
            if parseString2Pagenum(None, string)[-1] != newlist[i]:
                if dash == True:
                    string += "-"+str(newlist[i])
                else:
                    string += ", "+str(newlist[i])
    return string


def removewrongUTF8(name):
    newname = u""
    for char in name:
        try:
            codecs.decode(char, "UTF-8")
        except:
            pass
        else:
            newname += char
    return newname


def getMainIcon(pxlength=32):
    """ *pxlength* is the side length in pixels of the icon """
    # Set window icon
    iconBMP = icon.getMainBitmap()
    # scale
    image = wx.Bitmap.ConvertToImage(iconBMP)
    image = image.Scale(pxlength, pxlength, wx.IMAGE_QUALITY_HIGH)
    iconBMP = wx.Bitmap(image)
    iconICO = wx.Icon(iconBMP)
    return iconICO



"""Module frontend

The frontend displays the GUI (Graphic User Interface).
All functions and modules are called from here.
"""
import numpy as np
import warnings
import wx
import wx.lib.plot as plot
import wx.lib.scrolledpanel as scrolled


from pycorrfit import models as mdls
from pycorrfit import fit
from pycorrfit import Correlation


from . import tools
from . import wxutils


class FittingPanel(wx.Panel):
    """
    Those are the Panels that show the fitting dialogs with the Plots.
    """

    def __init__(self, parent, counter, modelid, active_parms, tau=None):
        """ Initialize with given parameters. """
        wx.Panel.__init__(self, parent=parent, id=wx.ID_ANY)
        self.parent = parent

        self.corr = Correlation(fit_model=modelid)
        if tau is not None:
            self.corr.lag_time = tau
        # active_parameters:
        # [0] labels
        # [1] values
        # [2] bool values to fit
        self.corr.fit_parameters = active_parms[1]
        self.corr.fit_parameters_variable = active_parms[2]

        self._bgselected = None
        self._bg2selected = None

        self.FitKnots = 5  # number of knots for spline fit or similiars

        self.weighted_fittype_id = 0  # integer (drop down item)
        self.weighted_nuvar = 3  # bins for std-dev. (left and rigth)

        # The weights that are plotted in the page
        # This is set by the PlotAll function
        self.weights_plot_fill_area = None

        # A list containing page numbers that share parameters with this page.
        # This parameter is defined by the global fitting tool and is saved in
        # sessions.
        self.GlobalParameterShare = []
        # Counts number of Pages already created:
        self.counter = counter
        # Has inital plot been performed?
        # Call PlotAll("init") to set this to true. If it is true, then
        # nothing will be plotted if called with "init"
        self.InitialPlot = False
        # Model we are using

        # Tool statistics uses this list:
        self.StatisticsCheckboxes = None
        # Splitter window
        # Sizes
        size = parent.notebook.GetSize()
        tabsize = 33
        size[1] = size[1] - tabsize
        self.sizepanelx = 270
        canvasx = size[0]-self.sizepanelx+5
        sizepanel = (self.sizepanelx, size[1])
        sizecanvas = (canvasx, size[1])
        self.sp = wx.SplitterWindow(self, size=size, style=wx.SP_3DSASH)
        # This is necessary to prevent "Unsplit" of the SplitterWindow:
        self.sp.SetMinimumPaneSize(1)
        # Settings Section (left side)
        # self.panelsettings = wx.Panel(self.sp, size=sizepanel)
        self.panelsettings = scrolled.ScrolledPanel(self.sp, size=sizepanel)
        self.panelsettings.SetupScrolling(scroll_x=False)
        # Setting up Plot (correlation + chi**2)
        self.spcanvas = wx.SplitterWindow(self.sp, size=sizecanvas,
                                          style=wx.SP_3DSASH)
        # This is necessary to prevent "Unsplit" of the SplitterWindow:
        self.spcanvas.SetMinimumPaneSize(1)
        # y difference in pixels between Auocorrelation and Residuals
        cupsizey = (size[1]*4)//5
        # Calculate initial data
        self.calculate_corr()
        # Draw the settings section
        self.settings()
        # Load default values
        self.apply_parameters_reverse()
        # Upper Plot for plotting of Correlation Function
        self.canvascorr = plot.PlotCanvas(self.spcanvas)
        self.canvascorr.logScale = (True, False)
        self.canvascorr.enableZoom = True
        self.PlotAll(event="init", trigger="tab_init")
        self.canvascorr.SetSize((canvasx, cupsizey))
        # Lower Plot for plotting of the residuals
        self.canvaserr = plot.PlotCanvas(self.spcanvas)
        self.canvaserr.logScale = (True, False)
        self.canvaserr.enableZoom = True
        self.canvaserr.SetSize((canvasx, size[1]-cupsizey))
        self.spcanvas.SplitHorizontally(self.canvascorr, self.canvaserr,
                                        cupsizey)
        self.sp.SplitVertically(self.panelsettings, self.spcanvas,
                                self.sizepanelx)
        # Bind resizing to resizing function.
        wx.EVT_SIZE(self, self.OnSize)

    @property
    def active_parms(self):
        names = self.corr.fit_model.parameters[0]
        parms = self.corr.fit_parameters
        bools = self.corr.fit_parameters_variable
        return [names, parms, bools]

    @property
    def IsCrossCorrelation(self):
        return self.corr.is_cc

    @property
    def modelid(self):
        return self.corr.fit_model.id

    @property
    def prevent_batch_modification(self):
        return self.wxCBPreventBatchParms.GetValue()

    @prevent_batch_modification.setter
    def prevent_batch_modification(self, value):
        self.wxCBPreventBatchParms.SetValue(value)

    @property
    def title(self):
        return self.tabtitle.GetValue()

    @title.setter
    def title(self, title):
        self.tabtitle.SetValue(title.strip())
        self.corr.title = title.strip()

    @property
    def traceavg(self):
        warnings.warn("Trace average always set to none!")
        return None

    @property
    def tracecc(self):
        if self.corr.is_cc and len(self.corr.traces) != 0:
            return self.corr.traces
        else:
            return None

    @property
    def bgselected(self):
        return self._bgselected

    @bgselected.setter
    def bgselected(self, value):
        if value is None:
            self.corr.backgrounds = []
            return
        # check paren.Background and get id
        background = self.parent.Background[value]
        self.corr.background_replace(0, background)
        self._bgselected = value

    @property
    def bg2selected(self):
        return self._bg2selected

    @bg2selected.setter
    def bg2selected(self, value):
        if value is None:
            if self.corr.is_cc:
                self.corr.backgrounds = []
            return
        # check paren.Background and get id
        background = self.parent.Background[value]
        self.corr.background_replace(1, background)
        self._bg2selected = value

    def apply_parameters(self, event=None):
        """ Read the values from the GUI form and write it to the
            pages parameters / correlation class.
            This function is called when the "Apply" button is hit.
        """
        modelid = self.corr.fit_model.id
        parameters = list()
        parameters_variable = list()
        # Read parameters from form and update self.active_parms[1]
        for i in np.arange(len(self.spincontrol)):
            parameters.append(1*self.spincontrol[i].GetValue())
            parameters_variable.append(self.checkboxes[i].GetValue())

        self.corr.fit_parameters_variable = np.array(parameters_variable,
                                                     dtype=bool)
        # As of version 0.7.5: we want the units to be displayed
        # human readable - the way they are displayed
        # in the Page info tool.
        # Here: Convert human readable units to program internal
        # units
        parmsconv = mdls.GetInternalFromHumanReadableParm(
            modelid, np.array(parameters))[1]
        self.corr.fit_parameters = parmsconv

        # Fitting parameters
        self.weighted_nuvar = self.Fitbox[5].GetValue()

        self.weighted_fittype_id = self.Fitbox[1].GetSelection()

        fitbox_value = self.Fitbox[1].GetValue()

        if self.weighted_fittype_id == -1:
            # User edited knot number
            Knots = fitbox_value
            Knots = "".join(filter(lambda x: x.isdigit(), Knots))
            if Knots == "":
                Knots = "5"
            self.weighted_fittype_id = 1
            self.FitKnots = str(Knots)
            fit_weight_type = "spline{}".format(self.FitKnots)
            fit_weight_data = self.weighted_nuvar
        elif self.weighted_fittype_id == 1:
            Knots = fitbox_value
            Knots = "".join(filter(lambda x: x.isdigit(), Knots))
            self.FitKnots = int(Knots)
            fit_weight_type = "spline{}".format(self.FitKnots)
            fit_weight_data = self.weighted_nuvar
        elif self.weighted_fittype_id == 0:
            fit_weight_type = "none"
            fit_weight_data = None
        elif self.weighted_fittype_id == 2:
            fit_weight_type = "model function"
            fit_weight_data = self.weighted_nuvar
        else:  # fitbox_selection > 2:
            fit_weight_type = fitbox_value
            self.corr.fit_weight_type = fitbox_value
            fit_weight_data = self.corr.fit_weight_data

        # Fitting algorithm
        keys = fit.GetAlgorithmStringList()[0]
        idalg = self.AlgorithmDropdown.GetSelection()

        self.corr.fit_algorithm = keys[idalg]
        self.corr.fit_weight_type = fit_weight_type
        self.corr.fit_weight_data = fit_weight_data

        # If parameters have been changed because of the check_parms
        # function, write them back.
        self.apply_parameters_reverse()

    def apply_parameters_reverse(self, event=None):
        """ Read the values from the pages parameters and write
            it to the GUI form.
        """
        modelid = self.corr.fit_model.id
        #
        # As of version 0.7.5: we want the units to be displayed
        # human readable - the way they are displayed
        # in the Page info tool.
        #
        # Here: Convert program internal units to
        # human readable units
        parameters = mdls.GetHumanReadableParms(modelid,
                                                self.corr.fit_parameters)[1]
        parameters_variable = self.corr.fit_parameters_variable
        # Write parameters to the form on the Page
        for i in np.arange(len(self.active_parms[1])):
            self.spincontrol[i].SetValue(parameters[i])
            self.checkboxes[i].SetValue(bool(parameters_variable[i]))
        # Fitting parameters
        self.Fitbox[5].SetValue(self.weighted_nuvar)
        idf = self.weighted_fittype_id
        List = self.Fitbox[1].GetItems()
        List[1] = "spline ("+str(self.FitKnots)+" knots)"
        self.Fitbox[1].SetItems(List)
        self.Fitbox[1].SetSelection(idf)
        # Normalization
        if self.corr.normparm is None:
            normsel = 0
        else:
            normsel = self.corr.normparm + 1
        self.AmplitudeInfo[2].SetSelection(normsel)
        # Fitting algorithm
        keys = fit.GetAlgorithmStringList()[0]
        idalg = keys.index(self.corr.fit_algorithm)
        self.AlgorithmDropdown.SetSelection(idalg)
        self.updateChi2()

    def calculate_corr(self):
        """
        Calculate model correlation function
        """
        return self.corr.modeled

    def Fit_enable_fitting(self):
        """ Enable the fitting button and the weighted fit control"""
        # self.Fitbox = [ fitbox, weightedfitdrop, fittext, fittext2,
        #                fittextvar, fitspin, buttonfit, textalg,
        #                self.AlgorithmDropdown]
        self.Fitbox[0].Enable()
        self.Fitbox[1].Enable()
        self.Fitbox[6].Enable()
        self.Fitbox[7].Enable()
        self.Fitbox[8].Enable()

    def Fit_function(self, event=None, noplots=False, trigger=None):
        """ Calls the fit function.

            `noplots=True` prevents plotting of spline fits

            `trigger` is passed to page.PlotAll.
                      If trigger is "fit_batch", then `noplots` is set
                      to `True`.

        """
        tools.batchcontrol.FitProgressDlg(self, self)

    def Fit_finalize(self, trigger):
        """ Things that need be done after fitting
        """
        # Reset list of globally shared parameters, because we are only
        # fitting this single page now.
        # TODO:
        # - also remove this page from the GlobalParameterShare list of
        #   the other pages
        self.GlobalParameterShare = []
        # Update spin-control values
        self.apply_parameters_reverse()
        # Plot everything
        try:
            self.PlotAll(trigger=trigger)
        except OverflowError:
            # Sometimes parameters are just bad and
            # we still want the user to use the
            # program.
            warnings.warn("Could not plot canvas.")
        # update displayed chi2
        self.updateChi2()

    def Fit_WeightedFitCheck(self, event=None):
        """ Enable Or disable variance calculation, dependent on
            "Weighted Fit" checkbox
        """
        # self.Fitbox=[ fitbox, weightedfitdrop, fittext, fittext2, fittextvar,
        #                fitspin, buttonfit ]
        weighted = (self.Fitbox[1].GetSelection() != 0)
        # In the case of "Average" we do not enable the
        # "Calculation of variance" part.
        if weighted is True and self.Fitbox[1].GetValue() != "Average":
            self.Fitbox[2].Enable()
            self.Fitbox[3].Enable()
            self.Fitbox[4].Enable()
            self.Fitbox[5].Enable()
        else:
            self.Fitbox[2].Disable()
            self.Fitbox[3].Disable()
            self.Fitbox[4].Disable()
            self.Fitbox[5].Disable()

    def MakeStaticBoxSizer(self, boxlabel):
        """ Create a Box with check boxes (fit yes/no) and possibilities to
            change initial values for fitting.

            Parameters:
            *boxlabel*: The name of the box (is being displayed)
            *self.active_parms[0]*: A list of things to put into the box

            Returns:
            *sizer*: The static Box
            *check*: The (un)set checkboxes
            *spin*: The spin text fields
        """
        modelid = self.corr.fit_model.id
        box = wx.StaticBox(self.panelsettings, label=boxlabel)
        sizer = wx.StaticBoxSizer(box, wx.VERTICAL)
        check = list()
        spin = list()
        #
        # As of version 0.7.5: we want the units to be displayed
        # human readable - the way they are displayed
        # in the Page info tool.
        #
        labels = mdls.GetHumanReadableParms(modelid,
                                            self.corr.fit_parameters)[0]
        sizerh = wx.GridSizer(len(labels), 2, 2, 2)
        for label in labels:
            checkbox = wx.CheckBox(self.panelsettings, label=label)
            # We needed to "from wx.lib.agw import floatspin" to get this:
            spinctrl = wxutils.PCFFloatTextCtrl(self.panelsettings)
            sizerh.Add(spinctrl)
            sizerh.Add(checkbox)
            # Put everything into lists to be able to refer to it later
            check.append(checkbox)
            spin.append(spinctrl)
        sizer.Add(sizerh)
        return sizer, check, spin

    def OnAmplitudeCheck(self, event=None):
        """ Enable/Disable BG rate text line.
            New feature introduced in 0.7.8
        """
        modelid = self.corr.fit_model.id
        # Normalization to a certain parameter in plots
        # Find all parameters that start with an "N"
        # ? and "C" ?
        # Create List
        normlist = list()
        normlist.append("None")
        # Add parameters
        parameterlist = list()
        for i in np.arange(len(self.active_parms[0])):
            label = self.active_parms[0][i]
            if label[0].lower() == "n":
                normlist.append("*"+label)
                parameterlist.append(i)
        # Add supplementary parameters
        # Get them from models
        supplement = mdls.GetMoreInfo(modelid, self)
        if supplement is not None:
            for i in np.arange(len(supplement)):
                label = supplement[i][0]
                if label[0].lower() == "n":
                    normlist.append("*"+label)
                    # Add the id of the supplement starting at the
                    # number of fitting parameters of current page.
                    parameterlist.append(i+len(self.active_parms[0]))
        normsel = self.AmplitudeInfo[2].GetSelection()
        if normsel in [0, -1]:
            # init or no normalization selected
            self.corr.normparm = None
            normsel = 0
        else:
            self.corr.normparm = parameterlist[normsel-1]

        if len(parameterlist) > 0:
            self.AmplitudeInfo[2].Enable()
            self.AmplitudeInfo[3].Enable()
        else:
            self.AmplitudeInfo[2].Disable()
            self.AmplitudeInfo[3].Disable()
        # Set dropdown values
        self.AmplitudeInfo[2].SetItems(normlist)
        self.AmplitudeInfo[2].SetSelection(normsel)
        # Plot intensities
        # Quick reminder:
        # self.AmplitudeInfo = [ [intlabel1, intlabel2],
        #                       [bgspin1, bgspin2],
        #                       normtoNDropdown, textnor]
        # Signal
        self.AmplitudeInfo[0][0].SetValue("{:.4f}".format(0))
        self.AmplitudeInfo[0][1].SetValue("{:.4f}".format(0))
        for i in range(len(self.corr.traces)):
            S = self.corr.traces[i].countrate
            self.AmplitudeInfo[0][i].SetValue("{:.4f}".format(S))
        if self.corr.is_cc:
            self.AmplitudeInfo[0][1].Enable()
        else:
            self.AmplitudeInfo[0][1].Disable()
        # Background
        # self.parent.Background[self.bgselected][i]
        # [0] average signal [kHz]
        # [1] signal name (edited by user)
        # [2] signal trace (tuple) ([ms], [kHz])
        if len(self.corr.backgrounds) >= 1:
            self.AmplitudeInfo[1][0].SetValue(
                self.corr.backgrounds[0].countrate)
        else:
            self.AmplitudeInfo[1][0].SetValue(0)
            self.AmplitudeInfo[1][1].SetValue(0)

        if len(self.corr.backgrounds) == 2:
            self.AmplitudeInfo[1][1].SetValue(
                self.corr.backgrounds[1].countrate)
        else:
            self.AmplitudeInfo[1][1].SetValue(0)
        # Disable the second line in amplitude correction, if we have
        # autocorrelation only.
        boolval = self.corr.is_cc
        for item in self.WXAmplitudeCCOnlyStuff:
            item.Enable(boolval)

    def OnBGSpinChanged(self, e):
        """ Calls tools.background.ApplyAutomaticBackground
            to update background information
        """
        # Quick reminder:
        # self.AmplitudeInfo = [ [intlabel1, intlabel2],
        #                       [bgspin1, bgspin2],
        #                       normtoNDropdown, textnor]
        if self.corr.is_cc:
            # update both self.bgselected and self.bg2selected
            bg = [float(self.AmplitudeInfo[1][0].GetValue()),
                  float(self.AmplitudeInfo[1][1].GetValue())]
            sig = [float(self.AmplitudeInfo[0][0].GetValue()),
                   float(self.AmplitudeInfo[0][1].GetValue())]
            # Make sure bg < sig
            for ii in range(len(bg)):
                if sig[ii] != 0:
                    if bg[ii] > .99*sig[ii]:
                        bg[ii] = .99*sig[ii]
                        self.AmplitudeInfo[1][ii].SetValue(bg[ii])
        else:
            # Only update self.bgselected
            bg = float(self.AmplitudeInfo[1][0].GetValue())
            sig = float(self.AmplitudeInfo[0][0].GetValue())
            # Make sure bg < sig
            if sig != 0:
                if bg > .99*sig:
                    bg = .99*sig
                    self.AmplitudeInfo[1][0].SetValue(bg)
        tools.background.ApplyAutomaticBackground(self, bg,
                                                  self.parent)
        e.Skip()

    def OnTitleChanged(self, e):
        modelid = self.corr.fit_model.id
        pid = self.parent.notebook.GetPageIndex(self)
        if self.tabtitle.GetValue() == "":
            text = self.counter + mdls.modeldict[modelid][1]
        else:
            # How many characters of the the page title should be displayed
            # in the tab? We choose 9: AC1-012 plus 2 whitespaces
            text = self.counter + self.tabtitle.GetValue()[-9:]
        self.parent.notebook.SetPageText(pid, text)

    def OnSetRange(self, e):
        """ Open a new Frame where the parameter range can be set.
            Rewrites self.parameter_range
            Parameter ranges are treated like parameters: They are saved in
            sessions and applied in batch mode.
        """
        # TODO:
        # - make range selector work with new class

        # We write a separate tool for that.
        # This tool does not show up in the Tools menu.
        if self.parent.RangeSelector is None:
            self.parent.RangeSelector = tools.RangeSelector(self)
            self.parent.RangeSelector.Bind(wx.EVT_CLOSE,
                                           self.parent.RangeSelector.OnClose)
        else:
            try:
                self.parent.RangeSelector.OnClose()
            except:
                pass
            self.parent.RangeSelector = None

    def OnSize(self, event):
        """ Resize the fitting Panel, when Window is resized. """
        size = self.parent.notebook.GetSize()
        tabsize = 33
        size[1] = size[1] - tabsize
        self.sp.SetSize(size)

    def PlotAll(self, event=None, trigger=None):
        """
        This function plots the whole correlation and residuals canvas.
        We do:
        - Channel selection
        - Background correction
        - Apply Parameters (separate function)
        - Drawing of plots

        The `event` is usually just an event from buttons or similar
        wx objects. It can be "init", then some initial plotting is
        done before the data is handled.

        The `trigger` is passed to `self.parent.OnFNBPageChanged` so
        that tools can update their content accordingly. For more
        information on triggers, have a look at the doctring of the
        `tools` submodule.
        """
        if event == "init":
            # We use this to have the page plotted at least once before
            # readout of parameters (e.g. startcrop, endcrop)
            # This is a performence tweak.
            if self.InitialPlot:
                return
            else:
                self.InitialPlot = True
        # Enable/Disable, set values frontend normalization
        self.OnAmplitudeCheck()
        # Apply parameters
        self.apply_parameters()
        # Calculate correlation function from parameters
        # Drawing of correlation plot
        # Plots corr.correlation_fit and the calcualted correlation function
        # self.datacorr into the upper canvas.
        # Create a line @ y=zero:
        zerostart = self.corr.lag_time_fit[0]
        zeroend = self.corr.lag_time_fit[-1]
        datazero = [[zerostart, 0], [zeroend, 0]]
        # Set plot colors
        width = 1
        colexp = "grey"
        colfit = "blue"
        colweight = "cyan"
        lines = list()
        linezero = plot.PolyLine(datazero, colour='orange', width=width)
        lines.append(linezero)
        if self.corr.correlation is not None:
            if self.corr.is_weighted_fit and \
               self.parent.MenuShowWeights.IsChecked():
                try:
                    weights = self.corr.fit_results["fit weights"]
                except:
                    weights = self.corr.fit_weight_data

                if isinstance(weights, np.ndarray):
                    # user might have selected a new weight type and
                    # presses apply, do not try to display weights

                    # if weights are from average or other, make sure that the
                    # dimensions are correct
                    if weights.shape[0] == self.corr.correlation.shape[0]:
                        weights = weights[self.corr.fit_ival[0]:self.corr.fit_ival[1]]

                    # perform some checks
                    if np.allclose(weights, np.ones_like(weights)):
                        weights = 0
                    elif weights.shape[0] != self.corr.modeled_fit.shape[0]:
                        # non-matching weigths
                        warnings.warn(
                            "Unmatching weights found. Probably from previous data set.")
                        weights = 0

                    # Add the weights to the graph.
                    # This is done by drawing two lines.
                    w = 1*self.corr.modeled_fit
                    w1 = 1*w
                    w2 = 1*w

                    w1[:, 1] = w[:, 1] + weights
                    w2[:, 1] = w[:, 1] - weights
                    # crop w1 and w2 if corr.correlation_fit does not include all
                    # data points.
                    if np.all(w[:, 0] == self.corr.correlation_fit[:, 0]):
                        pass
                    else:
                        raise ValueError(
                            "This should not have happened: size of weights is wrong.")
                    # Normalization with self.normfactor
                    w1[:, 1] *= self.corr.normalize_factor
                    w2[:, 1] *= self.corr.normalize_factor
                    self.weights_plot_fill_area = [w1, w2]
                    lineweight1 = plot.PolyLine(w1, legend='',
                                                colour=colweight, width=width)
                    lines.append(lineweight1)
                    lineweight2 = plot.PolyLine(w2, legend='',
                                                colour=colweight, width=width)
                    lines.append(lineweight2)
            else:
                self.weights_plot_fill_area = None

            # Plot Correlation curves
            # Plot both, experimental and calculated data
            # Normalization with self.normfactor, new feature in 0.7.8
            datacorr_norm = self.corr.modeled_plot
            linecorr = plot.PolyLine(datacorr_norm, legend='', colour=colfit,
                                     width=width)
            dataexp_norm = self.corr.correlation_plot
            lineexp = plot.PolyLine(dataexp_norm, legend='', colour=colexp,
                                    width=width)
            # Draw linezero first, so it is in the background
            lines.append(lineexp)
            lines.append(linecorr)
            PlotCorr = plot.PlotGraphics(lines,
                                         xLabel=u'lag time τ [ms]', yLabel=u'G(τ)')

            self.canvascorr.Draw(PlotCorr)
            # Calculate residuals
            resid_norm = self.corr.residuals_plot
            lineres = plot.PolyLine(resid_norm, legend='', colour=colfit,
                                    width=width)

            # residuals or weighted residuals?
            if self.corr.is_weighted_fit:
                yLabelRes = "weighted \nresiduals"
            else:
                yLabelRes = "residuals"
            PlotRes = plot.PlotGraphics([linezero, lineres],
                                        xLabel=u'lag time τ [ms]',
                                        yLabel=yLabelRes)
            self.canvaserr.Draw(PlotRes)
        else:
            # Amplitude normalization, new feature in 0.7.8
            datacorr_norm = self.corr.modeled_plot
            linecorr = plot.PolyLine(datacorr_norm, legend='', colour='blue',
                                     width=1)
            PlotCorr = plot.PlotGraphics([linezero, linecorr],
                                         xLabel=u'Lag time τ [ms]', yLabel=u'G(τ)')
            self.canvascorr.Draw(PlotCorr)
        self.Refresh()
        self.parent.OnFNBPageChanged(trigger=trigger)

    def settings(self):
        """ Here we define, what should be displayed at the left side
            of the fitting page/tab.
            Parameters:
        """
        modelid = self.corr.fit_model.id
        horizontalsize = self.sizepanelx-10
        # Title
        # Create empty tab title
        mddat = mdls.modeldict[modelid]
        modelshort = mdls.GetModelType(modelid)
        titlelabel = u"Data set {} ({} {})".format(
            self.counter.strip(" :"), modelshort, mddat[1])
        boxti = wx.StaticBox(self.panelsettings, label=titlelabel)
        sizerti = wx.StaticBoxSizer(boxti, wx.VERTICAL)
        sizerti.SetMinSize((horizontalsize, -1))
        self.tabtitle = wx.TextCtrl(self.panelsettings, value="",
                                    size=(horizontalsize-20, -1))
        self.Bind(wx.EVT_TEXT, self.OnTitleChanged, self.tabtitle)
        sizerti.Add(self.tabtitle, 0, wx.EXPAND, 0)
        # Create StaticBoxSizer
        box1, check, spin = self.MakeStaticBoxSizer("Model parameters")
        # Make the check boxes and spin-controls available everywhere
        self.checkboxes = check
        self.spincontrol = spin
        #
        # As of version 0.7.5: we want the units to be displayed
        # human readable - the way they are displayed
        # in the Page info tool.
        #
        labels, parameters = mdls.GetHumanReadableParms(modelid,
                                                        self.active_parms[1])
        parameterstofit = self.active_parms[2]
        # Set initial values given by user/programmer for Diffusion Model
        for i in np.arange(len(labels)):
            self.checkboxes[i].SetValue(bool(parameterstofit[i]))
            self.spincontrol[i].SetValue(parameters[i])
        # Put everything together
        self.panelsettings.sizer = wx.BoxSizer(wx.VERTICAL)
        self.panelsettings.sizer.Add(sizerti, 0, wx.EXPAND, 0)
        self.panelsettings.sizer.Add(box1, 0, wx.EXPAND, 0)
        # checkbox "Protect from batch control"
        self.wxCBPreventBatchParms = wx.CheckBox(self.panelsettings,
                                                 -1,
                                                 "prevent batch modification")
        box1.Add(self.wxCBPreventBatchParms)
        # buttons "Apply" and "Set range"
        horzs = wx.BoxSizer(wx.HORIZONTAL)
        buttonapply = wx.Button(self.panelsettings, label="Apply")
        self.Bind(wx.EVT_BUTTON, self.PlotAll, buttonapply)
        horzs.Add(buttonapply)
        buttonrange = wx.Button(self.panelsettings, label="Set range")
        self.Bind(wx.EVT_BUTTON, self.OnSetRange, buttonrange)
        horzs.Add(buttonrange)
        box1.Add(horzs)
        # Set horizontal size
        box1.SetMinSize((horizontalsize, -1))
        # More info
        normbox = wx.StaticBox(
            self.panelsettings, label="Amplitude corrections")
        miscsizer = wx.StaticBoxSizer(normbox, wx.VERTICAL)
        miscsizer.SetMinSize((horizontalsize, -1))
        # Intensities and Background
        sizeint = wx.FlexGridSizer(rows=3, cols=3, vgap=5, hgap=5)
        sizeint.Add(wx.StaticText(self.panelsettings, label="[kHz]"))
        sizeint.Add(wx.StaticText(self.panelsettings,
                                  label="Intensity"))
        sizeint.Add(wx.StaticText(self.panelsettings,
                                  label="Background"))
        sizeint.Add(wx.StaticText(self.panelsettings, label="Ch1"))
        intlabel1 = wx.TextCtrl(self.panelsettings)
        bgspin1 = wxutils.PCFFloatTextCtrl(self.panelsettings)
        bgspin1.Bind(wx.EVT_KILL_FOCUS, self.OnBGSpinChanged)
        bgspin1.Bind(wx.EVT_TEXT_ENTER, self.OnBGSpinChanged)
        sizeint.Add(intlabel1)
        intlabel1.SetEditable(False)
        sizeint.Add(bgspin1)
        chtext2 = wx.StaticText(self.panelsettings, label="Ch2")
        sizeint.Add(chtext2)
        intlabel2 = wx.TextCtrl(self.panelsettings)
        intlabel2.SetEditable(False)
        bgspin2 = wxutils.PCFFloatTextCtrl(self.panelsettings)
        bgspin2.Bind(wx.EVT_KILL_FOCUS, self.OnBGSpinChanged)
        sizeint.Add(intlabel2)
        sizeint.Add(bgspin2)
        miscsizer.Add(sizeint)
        # Normalize to n?
        textnor = wx.StaticText(self.panelsettings, label="Plot normalization")
        miscsizer.Add(textnor)
        normtoNDropdown = wx.ComboBox(self.panelsettings)
        self.Bind(wx.EVT_COMBOBOX, self.PlotAll, normtoNDropdown)
        miscsizer.Add(normtoNDropdown)
        self.AmplitudeInfo = [[intlabel1, intlabel2],
                              [bgspin1, bgspin2],
                              normtoNDropdown, textnor]
        self.WXAmplitudeCCOnlyStuff = [chtext2, intlabel2, bgspin2]
        self.panelsettings.sizer.Add(miscsizer, 0, wx.EXPAND, 0)
        # Add fitting Box
        fitbox = wx.StaticBox(self.panelsettings, label="Fitting options")
        fitsizer = wx.StaticBoxSizer(fitbox, wx.VERTICAL)
        fitsizer.SetMinSize((horizontalsize, -1))
        # Add a checkbox for weighted fitting
        weightedfitdrop = wx.ComboBox(self.panelsettings)
        self.weightlist = ["no weights", "spline (5 knots)", "model function"]
        weightedfitdrop.SetItems(self.weightlist)
        weightedfitdrop.SetSelection(0)
        fitsizer.Add(weightedfitdrop)
        # WeightedFitCheck() Enables or Disables the variance part
        weightedfitdrop.Bind(wx.EVT_COMBOBOX, self.Fit_WeightedFitCheck)
        # Add the variance part.
        # In order to do a weighted fit, we need to calculate the variance
        # at each point of the experimental data array.
        # In order to do that, we need to know how many data points from left
        # and right of the interesting data point we want to include in that
        # calculation.
        fittext = wx.StaticText(self.panelsettings,
                                label="Calculation of the variance")
        fitsizer.Add(fittext)
        fittext2 = wx.StaticText(self.panelsettings,
                                 label="from 2j+1 data points")
        fitsizer.Add(fittext2)
        fitsizerspin = wx.BoxSizer(wx.HORIZONTAL)
        fittextvar = wx.StaticText(self.panelsettings, label="j = ")
        fitspin = wx.SpinCtrl(self.panelsettings, -1,
                              initial=3, min=1, max=100)
        fitsizerspin.Add(fittextvar)
        fitsizerspin.Add(fitspin)
        fitsizer.Add(fitsizerspin)
        # Add algorithm selection
        textalg = wx.StaticText(self.panelsettings, label="Algorithm")
        fitsizer.Add(textalg)
        self.AlgorithmDropdown = wx.ComboBox(self.panelsettings)
        items = fit.GetAlgorithmStringList()[1]
        self.AlgorithmDropdown.SetItems(items)
        self.Bind(wx.EVT_COMBOBOX, self.apply_parameters,
                  self.AlgorithmDropdown)
        fitsizer.Add(self.AlgorithmDropdown)
        self.AlgorithmDropdown.SetMaxSize(weightedfitdrop.GetSize())
        # Add button "Fit" and chi2 display
        fitbuttonsizer = wx.BoxSizer(wx.HORIZONTAL)
        buttonfit = wx.Button(self.panelsettings, label="Fit")
        self.Bind(wx.EVT_BUTTON, self.Fit_function, buttonfit)
        fitbuttonsizer.Add(buttonfit)

        # add shortcut
        acctbl = wx.AcceleratorTable(
            [(wx.ACCEL_CTRL, ord('F'), buttonfit.GetId())])
        self.SetAcceleratorTable(acctbl)
        ##

        self.WXTextChi2 = wx.StaticText(self.panelsettings)
        # this StaticText is updated by `self.updateChi2()`
        fitbuttonsizer.Add(self.WXTextChi2, flag=wx.ALIGN_CENTER)
        fitsizer.Add(fitbuttonsizer)
        self.panelsettings.sizer.Add(fitsizer, 0, wx.EXPAND, 0)
        # Squeeze everything into the sizer
        self.panelsettings.SetSizer(self.panelsettings.sizer)
        # This is also necessary in Windows
        self.panelsettings.Layout()
        self.panelsettings.Show()
        # Make all the stuff available for everyone
        self.Fitbox = [fitbox, weightedfitdrop, fittext, fittext2,
                       fittextvar, fitspin, buttonfit, textalg,
                       self.AlgorithmDropdown]
        # Disable Fitting since no data has been loaded yet
        for element in self.Fitbox:
            element.Disable()
        self.panelsettings.sizer.Fit(self.panelsettings)
        self.parent.Layout()

    def updateChi2(self):
        """
        updates the self.WXTextChi2 text control
        """
        label = u""
        if hasattr(self.corr, "fit_results"):
            if "chi2" in self.corr.fit_results:
                chi2 = self.corr.fit_results["chi2"]
                chi2str = wxutils.float2string_nsf(chi2, n=3)
                chi2str = wxutils.nice_string(chi2str)
                label = u"  χ²={}".format(chi2str)
        # This does not work with wxPython 2.8.12:
        # self.WXTextChi2.SetLabelMarkup(u"<b>{}</b>".format(label))
        self.WXTextChi2.SetLabel(u"{}".format(label))



"""PicoQuant .cor files"""
import csv
import pathlib
import warnings

import numpy as np


class LoadCORError(BaseException):
    pass


def get_header_index(L, elem):
    try:
        index = L.index(elem)
    except ValueError:
        raise ValueError(
            f'Expected {elem} in the data header, but found only {L}')

    return index


def openCOR(path, filename=None):
    """
    Read data from a PicoQuant .cor file.

    This format consists of a metadata header, a blank line, then a
    whitespace-delimited data header followed by a data array.
    ------------------------------------------------------------
    TTTR Correlator Export
    PicoHarp Software version 3.0.0.3 format version 3.0
    Raw data: c:\\users\\baker lab 432-a\\desktop\\grant fcs\\default_013.ptu
    Recorded: 16/12/22 17:40:13
    Mode: T2
    Routing Mask A: 0 1 0 0 0
    Routing Mask B: 1 0 0 0 0
    Start time [s]: 0.000000
    Time span [s]: 7.545534
    Counts A: 382989
    Counts B: 523316
    Tau resolution [s]: 0.00000002500000

     taustep       tau/s        G(A,A)    G(B,B)    G(A,B)
          6     0.0000001500    0.8375    0.3556    0.1708
          7     0.0000001750    0.6503    0.2652    0.1445
          9     0.0000002250    0.5680    0.2173    0.1264
         11     0.0000002750    0.4836    0.1594    0.1614
         13     0.0000003250    0.3241    0.1528    0.1486
         15     0.0000003750    0.2819    0.1236    0.1641
         17     0.0000004250    0.3591    0.1467    0.1102
    [...]
    ------------------------------------------------------------

    Returns:
    A dictionary containing:
       Correlation: list of 2d np.array containing two columns:
              1st: tau in ms
              2nd: corresponding correlation signal
       Type: list of strings indictating the type of each correlation
       Filename: the basename for each file loaded
       Trace: blank list, since cor does not store intensity traces
    """
    path = pathlib.Path(path)
    if filename is not None:
        warnings.warn("Using `filename` is deprecated.", DeprecationWarning)
        path = path / filename
    filename = path.name

    s_to_ms = 1000  # cor provides seconds, we want ms

    corfile = path.open('r', encoding='utf-8')
    header = list()
    for line in corfile:
        header.append(line.strip())
        if header[-1] == '':
            break

    if header[0] != 'TTTR Correlator Export':
        raise ValueError(
            f'Error while reading {path.name}. Expected the first line to be "TTTR Correlator Export"')

    data_header = next(corfile).split()
    data = np.loadtxt(corfile)

    tau = data[:, get_header_index(data_header, 'tau/s')]*s_to_ms
    correlation_names = ['G(A,A)', 'G(A,B)', 'G(B,B)']
    correlation_types = ['AC',     'CC',     'AC']
    correlations = list()

    for correlation_name in correlation_names:
        correlation = data[:, get_header_index(data_header, correlation_name)]
        cor = np.zeros((len(correlation), 2), dtype=correlation.dtype)
        cor[:, 0] = tau
        cor[:, 1] = correlation
        correlations.append(cor)

    dictionary = dict()
    dictionary['Correlation'] = correlations
    dictionary['Trace'] = [[] for _ in dictionary['Correlation']]
    dictionary['Type'] = correlation_types
    dictionary['Filename'] = [path.name for _ in dictionary['Correlation']]
    return dictionary



import numpy as np

from .control import model_setup
from .cp_confocal import twod
from .cp_triplet import trip


# 2D simple gauss
def CF_Gxy_T_gauss(parms, tau):
    u""" Two-dimensional diffusion with a Gaussian laser profile,
        including a triplet component.
        The triplet factor takes into account a blinking term.
        Set *T* or *τ_trip* to 0, if no triplet component is wanted.

        triplet = 1 + T/(1-T)*exp(-τ/τ_trip)

        G(τ) = offset + 1/( n * (1+τ/τ_diff) )*triplet

        Calculation of diffusion coefficient and concentration
        from the effective radius of the detection profile (r₀ = 2*σ):
        D = r₀²/(4*τ_diff)
        Conc = n/(π*r₀²)

        *parms* - a list of parameters.
        Parameters (parms[i]):
        [0] n       Effective number of particles in confocal area
        [1] τ_diff  Characteristic residence time in confocal area
        [2] τ_trip  Characteristic residence time in triplet state
        [3] T       Fraction of particles in triplet (non-fluorescent) state
                    0 <= T < 1
        [4] offset
        *tau* - lag time
    """
    n = parms[0]
    taudiff = parms[1]
    tautrip = parms[2]
    T = parms[3]
    dc = parms[4]

    triplet = trip(tau=tau, tautrip=tautrip, T=T)

    BB = twod(tau=tau, taudiff=taudiff)

    G = dc + 1/n * BB * triplet
    return G


def supplements(parms, countrate=None):
    # We can only give you the effective particle number
    n = parms[0]
    Info = list()
    if countrate is not None:
        # CPP
        cpp = countrate/n
        Info.append(["cpp [kHz]", cpp])
    return Info


parms = [4.0, 0.4, 0.001, 0.01, 0.0]

# Boundaries
# strictly positive
boundaries = [[0, np.inf]]*len(parms)
# F
boundaries[3] = [0, .9999999999999]
boundaries[-1] = [-np.inf, np.inf]

model_setup(
    modelid=6002,
    name="2D diffusion with triplet (confocal)",
    comp="T+2D",
    mtype="Confocal (Gaussian) and triplet",
    fctn=CF_Gxy_T_gauss,
    par_labels=[
        u"n",
        u"τ_diff [ms]",
        u"τ_trip [ms]",
        u"T",
        u"offset"],
    par_values=parms,
    par_vary=[True, True, True, True, False],
    par_boundaries=boundaries,
    par_constraints=[[2, "<", 1]],
    par_hr_labels=[
        u"n",
        u"τ_diff [ms]",
        u"τ_trip [µs]",
        u"T",
        u"offset"],
    par_hr_factors=[1., 1., 1000., 1., 1.],
    supplementary_method=supplements
)



"""Session files"""
import os
import pathlib
import tempfile
import shutil

import numpy as np
import pytest

import data_file_dl
import pycorrfit as pcf

NOAPITOKEN = "GITHUB_API_TOKEN" not in os.environ

examplefile = "Zeiss_Confocor3_LSM780_FCCS_HeLa_2015/019_cp_KIND+BFA.fcs"


@pytest.mark.xfail(NOAPITOKEN, reason="Restrictions to GitHub API")
@pytest.mark.filterwarnings('ignore::pycorrfit.fit.StuckParameterWarning')
def test_basic():
    """This is a very rudimentary test for the session handling"""
    dfile = data_file_dl.get_data_file(examplefile)
    data = pcf.readfiles.open_any(dfile)
    corr = pcf.Correlation(correlation=data["Correlation"][0],
                           traces=data["Trace"][0],
                           corr_type=data["Type"][0],
                           filename=os.path.basename(dfile),
                           title="test correlation",
                           fit_model=6035  # confocal 3D+3D)
                           )
    corr.fit_parameters_variable = [True, True, True, True,
                                    False, False, False]
    # crop triplet data
    corr.fit_ival[0] = 8
    pcf.Fit(corr)

    tmpdir = tempfile.mkdtemp(prefix="pycorrfit_tests_")
    path = pathlib.Path(tmpdir) / "session.pcfs"

    fiterr = []
    for ii, fitpid in enumerate(corr.fit_results["fit parameters"]):
        fiterr.append([int(fitpid),
                       float(corr.fit_results["fit error estimation"][ii])])

    Infodict = {
        "Correlations": {
            1: [corr.lag_time, corr.correlation]},
        "Parameters": {
            1: ["#1:",
                corr.fit_model.id,
                corr.fit_parameters,
                corr.fit_parameters_variable,
                corr.fit_ival,
                [3, 3, 5, corr.fit_algorithm],
                [None, None],
                True,
                None,
                [[0.0, np.inf],
                 [0.0, np.inf],
                 [0.0, np.inf],
                 [0.0, 0.9999999999999],
                 [-np.inf, np.inf]]
                ]},
        "Supplements": {
            1: {"FitErr": fiterr,
                "Chi sq": float(corr.fit_results["chi2"]),
                "Global Share": [],
                }},
        "External Functions": {},
        "Traces": {},
        "Comments": {"Session": "No comment."},
        "Backgrounds": {},
        "External Weights": {},
        "Preferences": {},
    }

    pcf.openfile.SaveSessionData(sessionfile=str(path),
                                 Infodict=Infodict)

    ldt = pcf.openfile.LoadSessionData(str(path))

    # lag time only, shape (N,)
    assert np.allclose(data["Correlation"][0][:, 0], ldt["Correlations"][1][0])
    # lag time and correlation, shape (N, 2)
    assert np.allclose(corr.correlation, ldt["Correlations"][1][1])
    # parameters
    assert corr.fit_model.id == ldt["Parameters"][0][1]
    assert np.allclose(corr.fit_parameters, ldt["Parameters"][0][2])
    assert np.allclose(corr.fit_parameters_variable, ldt["Parameters"][0][3])
    assert np.allclose(corr.fit_ival, ldt["Parameters"][0][4])

    shutil.rmtree(tmpdir, ignore_errors=True)


if __name__ == "__main__":
    # Run all tests
    loc = locals()
    for key in list(loc.keys()):
        if key.startswith("test_") and hasattr(loc[key], "__call__"):
            loc[key]()



"""Module tools - RangeSelector

Select the range in which the parameter should reside for fitting.
"""
import numpy as np
import wx

from pycorrfit import models as mdls
from .. import wxutils


class RangeSelector(wx.Frame):
    # This tool is derived from a wx.frame.
    def __init__(self, Page):
        # parent is the main frame of PyCorrFit
        self.parent = Page.parent
        # Get the window positioning correctly
        pos = self.parent.GetPosition()
        pos = (pos[0]+100, pos[1]+100)
        wx.Frame.__init__(self, parent=self.parent, title="Parameter Range",
                          pos=pos, style=wx.DEFAULT_FRAME_STYLE | wx.FRAME_FLOAT_ON_PARENT)
        # Page - the currently active page of the notebook.
        self.Page = self.parent.notebook.GetCurrentPage()
        # Content
        self.panel = wx.Panel(self)
        self.topSizer = wx.BoxSizer(wx.VERTICAL)

        self.WXboxsizerlist = list()
        self.WXparmlist = list()
        self.OnPageChanged(self.Page)
        # Icon
        if self.parent.MainIcon is not None:
            wx.Frame.SetIcon(self, self.parent.MainIcon)
        self.Show(True)

    def FillPanel(self):
        """ Fill the panel with parameters from the page
        """
        corr = self.Page.corr
        self.parameter_range = np.zeros_like(corr.fit_parameters_range)
        leftbound = []
        for item in corr.fit_parameters_range[:, 0]:
            if item is None:
                leftbound.append(-np.inf)
            else:
                leftbound.append(item)
        labels, parmleft = mdls.GetHumanReadableParms(corr.fit_model.id,  # @UnusedVariable
                                                      leftbound)
        rightbound = []
        for item in corr.fit_parameters_range[:, 1]:
            if item is None:
                rightbound.append(np.inf)
            else:
                rightbound.append(item)

        labels, parmright = mdls.GetHumanReadableParms(corr.fit_model.id,
                                                       rightbound)
        self.parameter_range[:, 0] = np.array(parmleft)
        self.parameter_range[:, 1] = np.array(parmright)
        # create line

        self.WXboxsizer = wx.FlexGridSizer(
            rows=len(labels), cols=5, vgap=5, hgap=5)
        for i in range(len(labels)):
            left = wxutils.PCFFloatTextCtrl(self.panel)
            right = wxutils.PCFFloatTextCtrl(self.panel)
            left.SetValue(self.parameter_range[i][0])
            right.SetValue(self.parameter_range[i][1])
            left.Bind(wx.EVT_SPINCTRL, self.OnSetParmRange)
            right.Bind(wx.EVT_SPINCTRL, self.OnSetParmRange)
            text = wx.StaticText(self.panel, label=u'< ')
            text2 = wx.StaticText(self.panel, label=labels[i])
            text3 = wx.StaticText(self.panel, label=u' <')
            self.WXboxsizer.Add(left)
            self.WXboxsizer.Add(text)
            self.WXboxsizer.Add(text2, 0, wx.ALIGN_CENTER, 0)
            self.WXboxsizer.Add(text3)
            self.WXboxsizer.Add(right)
            self.WXparmlist.append([left, [text, text2, text3], right])
        self.WXboxsizer.Layout()

        self.topSizer.Add(self.WXboxsizer)
        self.btnapply = wx.Button(self.panel, wx.ID_ANY, 'Apply')
        self.Bind(wx.EVT_BUTTON, self.OnSetParmRange, self.btnapply)
        self.topSizer.Add(self.btnapply)

        self.topSizer.Layout()
        self.panel.SetSizer(self.topSizer)
        self.topSizer.Fit(self.panel)
        self.SetMinSize((self.topSizer.GetMinSize()[0] + 20,
                         self.topSizer.GetMinSize()[1] + 40))
        self.topSizer.Fit(self)

    def OnClose(self, event=None):
        # This is a necessary function for PyCorrFit.
        # Do not change it.
        self.parent.RangeSelector = None
        self.Destroy()

    def OnPageChanged(self, page=None, trigger=None):
        """
            This function is called, when something in the panel
            changes. The variable `trigger` is used to prevent this
            function from being executed to save stall time of the user.
            Forr a list of possible triggers, see the doc string of
            `tools`.
        """
        # When parent changes
        # This is a necessary function for PyCorrFit.
        # This is stuff that should be done when the active page
        # of the notebook changes.
        if trigger in ["parm_batch", "fit_batch", "page_add_batch"]:
            return
        self.Page = page
        if self.parent.notebook.GetPageCount() == 0:
            self.panel.Disable()
            return
        self.panel.Enable()
        try:
            self.btnapply.Destroy()
        except:
            pass
        for i in np.arange(len(self.WXparmlist)):
            self.WXparmlist[i][0].Destroy()  # start
            self.WXparmlist[i][1][0].Destroy()  # pname
            self.WXparmlist[i][1][1].Destroy()  # pname
            self.WXparmlist[i][2].Destroy()  # end
        del self.WXparmlist
        for i in np.arange(len(self.WXboxsizerlist)):
            self.WXboxsizer.Remove(0)
        self.WXboxsizerlist = list()
        self.WXparmlist = list()
        self.FillPanel()

    def OnSetParmRange(self, e):
        """ Called whenever something is edited in this frame.
            Writes back parameter ranges to the page
        """
        corr = self.Page.corr
        # Read out parameters from all controls
        for i in range(len(self.WXparmlist)):
            self.parameter_range[i][0] = self.WXparmlist[i][0].GetValue()
            self.parameter_range[i][1] = self.WXparmlist[i][2].GetValue()
            if self.parameter_range[i][0] > self.parameter_range[i][1]:
                self.parameter_range[i][1] = 1.01 * \
                    np.abs(self.parameter_range[i][0])
                self.WXparmlist[i][2].SetValue(self.parameter_range[i][1])
        # Set parameters
        parm0 = mdls.GetInternalFromHumanReadableParm(corr.fit_model.id,
                                                      self.parameter_range[:, 0])[1]
        parm1 = mdls.GetInternalFromHumanReadableParm(corr.fit_model.id,
                                                      self.parameter_range[:, 1])[1]
        corr.fit_parameters_range = np.dstack((parm0, parm1))[0]

        if self.parent.MenuAutocloseTools.IsChecked():
            # Autoclose
            self.OnClose()



"""Wrapper for Loading PicoQuant .pt3 data files

Wraps around FCS_point_correlator by Dominic Waithe
https://github.com/dwaithe/FCS_point_correlator
"""
import pathlib
import warnings

import numpy as np

from .read_pt3_scripts.correlation_objects import picoObject
from . import util


class ParameterClass():
    """Stores parameters for correlation """

    def __init__(self):
        # Where the data is stored.
        self.data = []
        self.objectRef = []
        self.subObjectRef = []
        self.colors = ['blue', 'green', 'red',
                       'cyan', 'magenta', 'yellow', 'black']
        self.numOfLoaded = 0
        self.NcascStart = 0
        self.NcascEnd = 25
        self.Nsub = 6
        self.winInt = 10
        self.photonCountBin = 25


def getTrace(picoObject, number):
    """
    Extracts trace `number` from a `picoObject`.

    Parameters
    ----------
    picoObject: instance of picoObject
        The data retreived from a pt3 file
    number:
        The id of the trace, can be 1 or 2.
    """

    attrint = "timeSeries{}".format(number)
    attrtime = "timeSeriesScale{}".format(number)

    # binned photon counts
    intensity = np.array(getattr(picoObject, attrint))
    # Time in ms for each bin
    time = np.array(getattr(picoObject, attrtime))
    # time delta
    deltat = np.abs(time[2]-time[1])

    trace = np.zeros((intensity.shape[0], 2))
    trace[:, 0] = time  # ms
    trace[:, 1] = intensity / deltat  # kHz

    # If the trace is too big. Wee need to bin it.
    newtrace = util.downsample_trace(trace)

    return newtrace


def openPT3(path, filename=None):
    """ Retreive correlation curves from PicoQuant data files 

    This function is a wrapper around the PicoQuant capability of
    FCS_Viewer by Dominic Waithe.
    """
    path = pathlib.Path(path)
    if filename is not None:
        warnings.warn("Using `filename` is deprecated.", DeprecationWarning)
        path = path / filename
    filename = path.name

    par_obj = ParameterClass()

    pt3file = picoObject(str(path), par_obj, None)

    po = pt3file

    auto = po.autoNorm
    # lag time [ms]
    autotime = po.autotime.reshape(-1)

    corrlist = list()
    typelist = list()
    tracelist = list()
    # Some data points are zero for some reason
    id1 = np.where(autotime != 0)

    # AC0 - autocorrelation CH0
    corrac0 = auto[:, 0, 0]
    if np.sum(np.abs(corrac0[id1])) != 0:
        typelist.append("AC0")
        # autotime,auto[:,0,0]
        corrlist.append(np.hstack((autotime[id1].reshape(-1, 1),
                                   corrac0[id1].reshape(-1, 1))))
        tracelist.append([getTrace(po, 1)])

    # AC1 - autocorrelation CH1
    corrac1 = auto[:, 1, 1]
    if np.sum(np.abs(corrac1[id1])) != 0:
        typelist.append("AC1")
        # autotime,auto[:,1,1]
        corrlist.append(np.hstack((autotime[id1].reshape(-1, 1),
                                   corrac1[id1].reshape(-1, 1))))
        tracelist.append([getTrace(po, 2)])

    # CC01 - Cross-Correlation CH0-CH1
    corrcc01 = auto[:, 0, 1]
    if np.sum(np.abs(corrcc01[id1])) != 0:
        typelist.append("CC01")
        # autotime,auto[:,0,1]
        corrlist.append(np.hstack((autotime[id1].reshape(-1, 1),
                                   corrcc01[id1].reshape(-1, 1))))
        tracelist.append([getTrace(po, 1), getTrace(po, 2)])

    # CC10 - Cross-Correlation CH1-CH0
    corrcc10 = auto[:, 1, 0]
    if np.sum(np.abs(corrcc10[id1])) != 0:
        typelist.append("CC10")
        # autotime,auto[:,1,0]
        corrlist.append(np.hstack((autotime[id1].reshape(-1, 1),
                                   corrcc10[id1].reshape(-1, 1))))
        tracelist.append([getTrace(po, 1), getTrace(po, 2)])

    filelist = [filename] * len(typelist)

    dictionary = dict()
    dictionary["Correlation"] = corrlist
    dictionary["Trace"] = tracelist
    dictionary["Type"] = typelist
    dictionary["Filename"] = filelist

    return dictionary



import numpy as np

from .control import model_setup
from .cp_confocal import threed
from .cp_triplet import trip


def CF_Gxyz_blink(parms, tau):
    u""" Three-dimanesional free diffusion with a Gaussian laser profile
        (eliptical), including a triplet component.
        The triplet factor takes into account a blinking term.
        Set *T* or *τ_trip* to 0, if no triplet component is wanted.

        G(τ) = offset + 1/( n*(1+τ/τ_diff) * sqrt(1 + τ/(SP²*τ_diff)) )
                    * ( 1+T/(1.-T)*exp(-τ/τ_trip) )

        Calculation of diffusion coefficient and concentration
        from the effective radius of the detection profile (r₀ = 2*σ):
        D = r₀²/(4*τ_diff)
        Conc = n/( sqrt(π³)*r₀²*z₀ )

        *parms* - a list of parameters.
        Parameters (parms[i]):
        [0] n       Effective number of particles in confocal volume
        [1] T       Fraction of particles in triplet (non-fluorescent) state
                    0 <= T < 1
        [2] τ_trip  Characteristic residence time in triplet state
        [3] τ_diff  Characteristic residence time in confocal volume
        [4] SP      SP=z₀/r₀ Structural parameter,
                    describes the axis ratio of the confocal volume
        [5] offset
        *tau* - lag time
    """
    n = parms[0]
    T = parms[1]
    tautrip = parms[2]
    taudiff = parms[3]
    SP = parms[4]
    off = parms[5]

    AA = trip(tau, tautrip, T)
    BB = threed(tau, taudiff, SP)

    G = off + 1/n * AA * BB
    return G


def supplements(parms, countrate=None):
    # We can only give you the effective particle number
    n = parms[0]
    Info = list()
    if countrate is not None:
        # CPP
        cpp = countrate/n
        Info.append(["cpp [kHz]", cpp])
    return Info


parms = [4.0, 0.2, 0.001, 0.4, 5.0, 0.0]

# Boundaries
boundaries = [[0, np.inf]]*len(parms)
# T
boundaries[1] = [0, .9999999999999]
boundaries[-1] = [-np.inf, np.inf]


model_setup(
    modelid=6011,
    name="3D diffusion with triplet (confocal)",
    comp="T+3D",
    mtype="Confocal (Gaussian) and triplet",
    fctn=CF_Gxyz_blink,
    par_labels=[
        u"n",
        u"T",
        u"τ_trip [ms]",
        u"τ_diff [ms]",
        u"SP",
        u"offset"],
    par_values=parms,
    par_vary=[True, True, True, True, False, False],
    par_boundaries=boundaries,
    par_constraints=[[3, ">", 2]],
    par_hr_labels=[
        u"n",
        u"T",
        u"τ_trip [µs]",
        u"τ_diff [ms]",
        u"SP",
        u"offset"],
    par_hr_factors=[1., 1., 1000., 1., 1., 1.],
    supplementary_method=supplements
)



import numpy as np

from .control import model_setup
from .cp_confocal import twod

# 2D simple gauss


def CF_Gxy_gauss(parms, tau):
    u""" Two-dimensional diffusion with a Gaussian laser profile.

        G(τ) = offset + 1/( n * (1+τ/τ_diff) )

        Calculation of diffusion coefficient and concentration
        from the effective radius of the detection profile (r₀ = 2*σ):
        D = r₀²/(4*τ_diff)
        Conc = n/(π*r₀²)

        *parms* - a list of parameters.
        Parameters (parms[i]):
        [0] n       Effective number of particles in confocal area
        [1] τ_diff  Characteristic residence time in confocal area
        [2] offset
        *tau* - lag time
    """
    n = parms[0]
    taudiff = parms[1]
    dc = parms[2]

    BB = twod(tau=tau, taudiff=taudiff)

    G = dc + 1/n * BB
    return G


def supplements(parms, countrate=None):
    # We can only give you the effective particle number
    n = parms[0]
    Info = list()
    if countrate is not None:
        # CPP
        cpp = countrate/n
        Info.append(["cpp [kHz]", cpp])
    return Info


parms = [4.0, 0.4, 0.0]

# boundaries
boundaries = [[0, np.inf]]*len(parms)
boundaries[-1] = [-np.inf, np.inf]

model_setup(
    modelid=6001,
    name="2D diffusion (confocal)",
    comp="2D",
    mtype="Confocal (Gaussian)",
    fctn=CF_Gxy_gauss,
    par_labels=[u"n",
                u"τ_diff [ms]",
                u"offset"],
    par_values=parms,
    par_vary=[True, True, False],
    par_boundaries=boundaries,
    supplementary_method=supplements
)



import numpy as np
import os
import sys
from .correlation_methods import *
from .import_methods import *
import time
from .fitting_methods import equation_
# from lmfit import minimize, Parameters,report_fit,report_errors, fit_report
import csv
import copy


"""FCS Bulk Correlation Software

    Copyright (C) 2015  Dominic Waithe

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along
    with this program; if not, write to the Free Software Foundation, Inc.,
    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""


class picoObject():
    # This is the class which holds the .pt3 data and parameters
    def __init__(self, filepath, par_obj, fit_obj):

        # parameter object and fit object. If
        self.par_obj = par_obj
        self.fit_obj = fit_obj
        self.type = 'mainObject'

        # self.PIE = 0
        self.filepath = str(filepath)
        self.nameAndExt = os.path.basename(self.filepath).split('.')
        self.name = self.nameAndExt[0]
        self.ext = self.nameAndExt[-1]

        self.par_obj.data.append(filepath)
        self.par_obj.objectRef.append(self)

        # Imports pt3 file format to object.
        self.unqID = self.par_obj.numOfLoaded

        # For fitting.
        self.objId1 = None
        self.objId2 = None
        self.objId3 = None
        self.objId4 = None
        self.processData()

        self.plotOn = True

    def processData(self):

        self.NcascStart = self.par_obj.NcascStart
        self.NcascEnd = self.par_obj.NcascEnd
        self.Nsub = self.par_obj.Nsub
        self.winInt = self.par_obj.winInt
        self.photonCountBin = self.par_obj.photonCountBin

        # File import

        if self.ext == 'pt3':
            self.subChanArr, self.trueTimeArr, self.dTimeArr, self.resolution = pt3import(
                self.filepath)
        if self.ext == 'ptu':
            self.subChanArr, self.trueTimeArr, self.dTimeArr, self.resolution = ptuimport(
                self.filepath)
        if self.ext == 'csv':
            self.subChanArr, self.trueTimeArr, self.dTimeArr, self.resolution = csvimport(
                self.filepath)
            # If the file is empty.
            if self.subChanArr == None:
                # Undoes any preparation of resource.
                self.par_obj.data.pop(-1)
                self.par_obj.objectRef.pop(-1)
                self.exit = True
                self.par_obj.image_status_text.showMessage(
                    "Your sample is not in the correct format.")
                self.par_obj.fit_obj.app.processEvents()
                return

        # Colour assigned to file.
        self.color = self.par_obj.colors[self.unqID % len(self.par_obj.colors)]

        # How many channels there are in the files.
        # Minus 1 because not interested in channel 15.
        self.numOfCH = np.unique(np.array(self.subChanArr)).__len__()-1

        # Finds the numbers which address the channels.
        self.ch_present = np.unique(np.array(self.subChanArr[0:100]))

        # Calculates decay function for both channels.
        self.photonDecayCh1, self.decayScale1 = delayTime2bin(np.array(
            self.dTimeArr), np.array(self.subChanArr), self.ch_present[0], self.winInt)

        if self.numOfCH == 2:
            self.photonDecayCh2, self.decayScale2 = delayTime2bin(np.array(
                self.dTimeArr), np.array(self.subChanArr), self.ch_present[1], self.winInt)

        # Time series of photon counts. For visualisation.
        self.timeSeries1, self.timeSeriesScale1 = delayTime2bin(np.array(
            self.trueTimeArr)/1000000, np.array(self.subChanArr), self.ch_present[0], self.photonCountBin)

        unit = self.timeSeriesScale1[-1]/self.timeSeriesScale1.__len__()

        # Converts to counts per
        self.kcount_CH1 = np.average(self.timeSeries1)

        # This is the unnormalised intensity count for int_time duration (the first moment)
        raw_count = np.average(self.timeSeries1)
        var_count = np.var(self.timeSeries1)

        self.brightnessNandBCH0 = (
            ((var_count - raw_count)/(raw_count))/(float(unit)))
        if (var_count-raw_count) == 0:
            self.numberNandBCH0 = 0
        else:
            self.numberNandBCH0 = (raw_count**2/(var_count-raw_count))

        if self.numOfCH == 2:

            self.timeSeries2, self.timeSeriesScale2 = delayTime2bin(np.array(
                self.trueTimeArr)/1000000, np.array(self.subChanArr), self.ch_present[1], self.photonCountBin)
            unit = self.timeSeriesScale2[-1]/self.timeSeriesScale2.__len__()
            self.kcount_CH2 = np.average(self.timeSeries2)
            # This is the unnormalised intensity count for int_time duration (the first moment)
            raw_count = np.average(self.timeSeries2)
            var_count = np.var(self.timeSeries2)
            self.brightnessNandBCH1 = (
                ((var_count - raw_count)/(raw_count))/(float(unit)))
            if (var_count-raw_count) == 0:
                self.numberNandBCH1 = 0
            else:
                self.numberNandBCH1 = (raw_count**2/(var_count-raw_count))

        # Calculates the Auto and Cross-correlation functions.
        self.crossAndAuto(np.array(self.trueTimeArr),
                          np.array(self.subChanArr))

        if self.fit_obj != None:
            # If fit object provided then creates fit objects.
            if self.objId1 == None:
                corrObj = corrObject(self.filepath, self.fit_obj)
                self.objId1 = corrObj.objId
                self.fit_obj.objIdArr.append(corrObj.objId)
                self.objId1.param = copy.deepcopy(self.fit_obj.def_param)
                self.objId1.name = self.name+'_CH0_Auto_Corr'
                self.objId1.ch_type = 0  # channel 0 Auto
                self.objId1.siblings = None
                self.objId1.prepare_for_fit()
                self.objId1.kcount = self.kcount_CH1

            self.objId1.autoNorm = np.array(self.autoNorm[:, 0, 0]).reshape(-1)
            self.objId1.autotime = np.array(self.autotime).reshape(-1)
            self.objId1.param = copy.deepcopy(self.fit_obj.def_param)

            if self.numOfCH == 2:
                if self.objId3 == None:
                    corrObj = corrObject(self.filepath, self.fit_obj)
                    self.objId3 = corrObj.objId
                    self.objId3.param = copy.deepcopy(self.fit_obj.def_param)
                    self.fit_obj.objIdArr.append(corrObj.objId)
                    self.objId3.name = self.name+'_CH1_Auto_Corr'
                    self.objId3.ch_type = 1  # channel 1 Auto
                    self.objId3.siblings = None
                    self.objId3.prepare_for_fit()
                    self.objId3.kcount = self.kcount_CH2

                self.objId3.autoNorm = np.array(
                    self.autoNorm[:, 1, 1]).reshape(-1)
                self.objId3.autotime = np.array(self.autotime).reshape(-1)
                self.objId3.param = copy.deepcopy(self.fit_obj.def_param)

                if self.objId2 == None:
                    corrObj = corrObject(self.filepath, self.fit_obj)
                    self.objId2 = corrObj.objId
                    self.objId2.param = copy.deepcopy(self.fit_obj.def_param)
                    self.fit_obj.objIdArr.append(corrObj.objId)
                    self.objId2.name = self.name+'_CH01_Cross_Corr'
                    self.objId2.ch_type = 2  # 01cross
                    self.objId2.siblings = None
                    self.objId2.prepare_for_fit()

                self.objId2.autoNorm = np.array(
                    self.autoNorm[:, 0, 1]).reshape(-1)
                self.objId2.autotime = np.array(self.autotime).reshape(-1)
                self.objId2.param = copy.deepcopy(self.fit_obj.def_param)

                if self.objId4 == None:
                    corrObj = corrObject(self.filepath, self.fit_obj)
                    self.objId4 = corrObj.objId
                    self.objId4.param = copy.deepcopy(self.fit_obj.def_param)
                    self.fit_obj.objIdArr.append(corrObj.objId)
                    self.objId4.name = self.name+'_CH10_Cross_Corr'
                    self.objId4.ch_type = 3  # 10cross
                    self.objId4.siblings = None
                    self.objId4.prepare_for_fit()

                self.objId4.autoNorm = np.array(
                    self.autoNorm[:, 1, 0]).reshape(-1)
                self.objId4.autotime = np.array(self.autotime).reshape(-1)
                self.objId4.param = copy.deepcopy(self.fit_obj.def_param)

            self.fit_obj.fill_series_list()
        self.dTimeMin = 0
        self.dTimeMax = np.max(self.dTimeArr)
        self.subDTimeMin = self.dTimeMin
        self.subDTimeMax = self.dTimeMax
        self.exit = False
        del self.subChanArr
        del self.trueTimeArr
        del self.dTimeArr

    def crossAndAuto(self, trueTimeArr, subChanArr):
        # For each channel we loop through and find only those in the correct time gate.
        # We only want photons in channel 1 or two.
        y = trueTimeArr[subChanArr < 3]
        validPhotons = subChanArr[subChanArr < 3]

        # Creates boolean for photon events in either channel.
        num = np.zeros((validPhotons.shape[0], 2))
        num[:, 0] = (np.array([np.array(validPhotons) ==
                               self.ch_present[0]])).astype(np.int32)
        if self.numOfCH == 2:
            num[:, 1] = (np.array([np.array(validPhotons) ==
                                   self.ch_present[1]])).astype(np.int32)

        self.count0 = np.sum(num[:, 0])
        self.count1 = np.sum(num[:, 1])

        t1 = time.time()
        auto, self.autotime = tttr2xfcs(
            y, num, self.NcascStart, self.NcascEnd, self.Nsub)
        t2 = time.time()

        # Normalisation of the TCSPC data:
        maxY = np.ceil(max(self.trueTimeArr))
        self.autoNorm = np.zeros((auto.shape))
        self.autoNorm[:, 0, 0] = (
            (auto[:, 0, 0]*maxY)/(self.count0*self.count0))-1

        if self.numOfCH == 2:
            self.autoNorm[:, 1, 1] = (
                (auto[:, 1, 1]*maxY)/(self.count1*self.count1))-1
            self.autoNorm[:, 1, 0] = (
                (auto[:, 1, 0]*maxY)/(self.count1*self.count0))-1
            self.autoNorm[:, 0, 1] = (
                (auto[:, 0, 1]*maxY)/(self.count0*self.count1))-1

        # Normalisaation of the decay functions.
        self.photonDecayCh1Min = np.array(self.photonDecayCh1) - \
            (np.min(self.photonDecayCh1) if len(self.photonDecayCh1) > 0 else 0)
        self.photonDecayCh1Norm = self.photonDecayCh1Min / \
            (np.max(self.photonDecayCh1Min) if len(self.photonDecayCh1) > 0 else 1)

        if self.numOfCH == 2:
            self.photonDecayCh2Min = np.array(self.photonDecayCh2) - \
                np.min(self.photonDecayCh2)
            self.photonDecayCh2Norm = self.photonDecayCh2Min / \
                np.max(self.photonDecayCh2Min)

        return


class subPicoObject():
    def __init__(self, parentId, xmin, xmax, TGid, par_obj):
        # Binning window for decay function
        self.TGid = TGid
        # Parameters for auto-correlation and cross-correlation.
        self.parentId = parentId
        self.par_obj = par_obj
        self.NcascStart = self.parentId.NcascStart
        self.NcascEnd = self.parentId.NcascEnd
        self.Nsub = self.parentId.Nsub
        self.fit_obj = self.parentId.fit_obj
        self.ext = self.parentId.ext

        self.type = 'subObject'
        # Appends the object to the subObject register.
        self.par_obj.subObjectRef.append(self)
        self.unqID = self.par_obj.subNum
        self.parentUnqID = self.parentId.unqID
        # self.chanArr = parentObj.chanArr
        # self.trueTimeArr = self.parentId.trueTimeArr
        # self.dTimeArr = self.parentId.dTimeArr
        self.color = self.parentId.color
        self.numOfCH = self.parentId.numOfCH
        self.ch_present = self.parentId.ch_present
        self.photonCountBin = self.par_obj.photonCountBin

        self.filepath = str(self.parentId.filepath)
        self.xmin = xmin
        self.xmax = xmax

        self.nameAndExt = os.path.basename(self.filepath).split('.')
        self.name = 'TG-'+str(self.unqID)+'-xmin_'+str(round(xmin, 0)) + \
            '-xmax_'+str(round(xmax, 0))+'-'+self.nameAndExt[0]

        self.objId1 = None
        self.objId2 = None
        self.objId3 = None
        self.objId4 = None
        self.processData()
        self.plotOn = True

    def processData(self):
        self.NcascStart = self.par_obj.NcascStart
        self.NcascEnd = self.par_obj.NcascEnd
        self.Nsub = self.par_obj.Nsub
        self.winInt = self.par_obj.winInt

        # self.subChanArr, self.trueTimeArr, self.dTimeArr,self.resolution = pt3import(self.filepath)
        if self.ext == 'pt3':
            self.subChanArr, self.trueTimeArr, self.dTimeArr, self.resolution = pt3import(
                self.filepath)
        if self.ext == 'csv':
            self.subChanArr, self.trueTimeArr, self.dTimeArr, self.resolution = csvimport(
                self.filepath)
            # If the file is empty.
            # if self.subChanArr == None:
            # Undoes any preparation of resource.
            #    self.par_obj.subObjectRef.pop(-1)
            # self.exit = True
            #    return

        self.subArrayGeneration(self.xmin, self.xmax)

        self.dTimeMin = self.parentId.dTimeMin
        self.dTimeMax = self.parentId.dTimeMax
        self.subDTimeMin = self.dTimeMin
        self.subDTimeMax = self.dTimeMax

       # Time series of photon counts. For visualisation.
        self.timeSeries1, self.timeSeriesScale1 = delayTime2bin(np.array(
            self.trueTimeArr)/1000000, np.array(self.subChanArr), self.ch_present[0], self.photonCountBin)

        unit = self.timeSeriesScale1[-1]/self.timeSeriesScale1.__len__()
        self.kcount_CH1 = np.average(self.timeSeries1)
        if self.numOfCH == 2:

            self.timeSeries2, self.timeSeriesScale2 = delayTime2bin(np.array(
                self.trueTimeArr)/1000000, np.array(self.subChanArr), self.ch_present[1], self.photonCountBin)
            unit = self.timeSeriesScale2[-1]/self.timeSeriesScale2.__len__()
            self.kcount_CH2 = np.average(self.timeSeries2)

        # Adds names to the fit function for later fitting.
        if self.objId1 == None:
            corrObj = corrObject(self.filepath, self.fit_obj)
            self.objId1 = corrObj.objId
            self.fit_obj.objIdArr.append(corrObj.objId)
            self.objId1.param = copy.deepcopy(self.fit_obj.def_param)
            self.objId1.name = self.name+'_CH0_Auto_Corr'
            self.objId1.ch_type = 0  # channel 0 Auto
            self.objId1.siblings = None
            self.objId1.prepare_for_fit()

            self.objId1.kcount = self.kcount_CH1
        self.objId1.autoNorm = np.array(self.autoNorm[:, 0, 0]).reshape(-1)
        self.objId1.autotime = np.array(self.autotime).reshape(-1)
        self.objId1.param = copy.deepcopy(self.fit_obj.def_param)

        if self.numOfCH == 2:
            if self.objId3 == None:
                corrObj = corrObject(self.filepath, self.fit_obj)
                self.objId3 = corrObj.objId
                self.fit_obj.objIdArr.append(corrObj.objId)
                self.objId3.param = copy.deepcopy(self.fit_obj.def_param)
                self.objId3.name = self.name+'_CH1_Auto_Corr'
                self.objId3.ch_type = 1  # channel 1 Auto
                self.objId3.siblings = None
                self.objId3.prepare_for_fit()
                self.objId3.kcount = self.kcount_CH2

            self.objId3.autoNorm = np.array(self.autoNorm[:, 1, 1]).reshape(-1)
            self.objId3.autotime = np.array(self.autotime).reshape(-1)
            self.objId3.param = copy.deepcopy(self.fit_obj.def_param)
            if self.objId2 == None:
                corrObj = corrObject(self.filepath, self.fit_obj)
                self.objId2 = corrObj.objId
                self.objId2.param = copy.deepcopy(self.fit_obj.def_param)
                self.fit_obj.objIdArr.append(corrObj.objId)
                self.objId2.name = self.name+'_CH01_Cross_Corr'
                self.objId2.ch_type = 2  # channel 01 Cross
                self.objId2.siblings = None
                self.objId2.prepare_for_fit()

            self.objId2.autoNorm = np.array(self.autoNorm[:, 0, 1]).reshape(-1)
            self.objId2.autotime = np.array(self.autotime).reshape(-1)
            self.objId2.param = copy.deepcopy(self.fit_obj.def_param)
            if self.objId4 == None:
                corrObj = corrObject(self.filepath, self.fit_obj)
                self.objId4 = corrObj.objId
                self.objId4.param = copy.deepcopy(self.fit_obj.def_param)
                self.fit_obj.objIdArr.append(corrObj.objId)
                self.objId4.name = self.name+'_CH10_Cross_Corr'
                self.objId4.ch_type = 3  # channel 10 Cross
                self.objId4.siblings = None
                self.objId4.prepare_for_fit()

            self.objId4.autoNorm = np.array(self.autoNorm[:, 1, 0]).reshape(-1)
            self.objId4.autotime = np.array(self.autotime).reshape(-1)

        self.fit_obj.fill_series_list()
        del self.subChanArr
        del self.trueTimeArr
        del self.dTimeArr

    def subArrayGeneration(self, xmin, xmax):
        if (xmax < xmin):
            xmin1 = xmin
            xmin = xmax
            xmax = xmin1
        # self.subChanArr = np.array(self.chanArr)
        # Finds those photons which arrive above certain time or below certain time.
        photonInd = np.logical_and(
            self.dTimeArr >= xmin, self.dTimeArr <= xmax).astype(bool)

        self.subChanArr[np.invert(photonInd).astype(bool)] = 16

        self.crossAndAuto()

        return

    def crossAndAuto(self):
        # We only want photons in channel 1 or two.
        validPhotons = self.subChanArr[self.subChanArr < 3]
        y = self.trueTimeArr[self.subChanArr < 3]
        # Creates boolean for photon events in either channel.
        num = np.zeros((validPhotons.shape[0], 2))
        num[:, 0] = (np.array([np.array(validPhotons) ==
                               self.ch_present[0]])).astype(np.int)
        if self.numOfCH == 2:
            num[:, 1] = (np.array([np.array(validPhotons) ==
                                   self.ch_present[1]])).astype(np.int)

        self.count0 = np.sum(num[:, 0])
        self.count1 = np.sum(num[:, 1])
        # Function which calculates auto-correlation and cross-correlation.

        auto, self.autotime = tttr2xfcs(
            y, num, self.NcascStart, self.NcascEnd, self.Nsub)

        maxY = np.ceil(max(self.trueTimeArr))
        self.autoNorm = np.zeros((auto.shape))
        self.autoNorm[:, 0, 0] = (
            (auto[:, 0, 0]*maxY)/(self.count0*self.count0))-1
        if self.numOfCH == 2:
            self.autoNorm[:, 1, 1] = (
                (auto[:, 1, 1]*maxY)/(self.count1*self.count1))-1
            self.autoNorm[:, 1, 0] = (
                (auto[:, 1, 0]*maxY)/(self.count1*self.count0))-1
            self.autoNorm[:, 0, 1] = (
                (auto[:, 0, 1]*maxY)/(self.count0*self.count1))-1

        return


class corrObject():
    def __init__(self, filepath, parentFn):
        # the container for the object.
        self.parentFn = parentFn
        self.type = 'corrObject'
        self.filepath = str(filepath)
        self.nameAndExt = os.path.basename(self.filepath).split('.')
        self.name = self.nameAndExt[0]
        self.ext = self.nameAndExt[-1]
        self.autoNorm = []
        self.autotime = []
        self.model_autoNorm = []
        self.model_autotime = []
        self.datalen = []
        self.objId = self
        self.param = []
        self.goodFit = True
        self.fitted = False
        self.checked = False
        self.clicked = False
        self.toFit = False
        self.kcount = None
        self.filter = False

    def prepare_for_fit(self):
        if self.parentFn.ch_check_ch0.isChecked() == True and self.ch_type == 0:
            self.toFit = True
        if self.parentFn.ch_check_ch1.isChecked() == True and self.ch_type == 1:
            self.toFit = True

        if self.parentFn.ch_check_ch01.isChecked() == True and self.ch_type == 2:
            self.toFit = True
        if self.parentFn.ch_check_ch10.isChecked() == True and self.ch_type == 3:
            self.toFit = True
        # self.parentFn.modelFitSel.clear()
        # for objId in self.parentFn.objIdArr:
         #   if objId.toFit == True:
          #      self.parentFn.modelFitSel.addItem(objId.name)
        self.parentFn.updateFitList()

    def residual(self, param, x, data, options):

        A = equation_(param, x, options)
        residuals = data-A
        return residuals

    def fitToParameters(self):
        # self.parentFn.updateParamFirst()
        # self.parentFn.updateTableFirst()
        # self.parentFn.updateParamFirst()

        # Populate param for lmfit.
        param = Parameters()
        # self.def_param.add('A1', value=1.0, min=0,max=1.0, vary=False)
        for art in self.param:

            if self.param[art]['to_show'] == True:

                param.add(art, value=float(self.param[art]['value']), min=float(
                    self.param[art]['minv']), max=float(self.param[art]['maxv']), vary=self.param[art]['vary'])

        # Find the index of the nearest point in the scale.

        data = np.array(self.autoNorm).astype(np.float64).reshape(-1)
        scale = np.array(self.autotime).astype(np.float64).reshape(-1)
        self.indx_L = int(np.argmin(np.abs(scale - self.parentFn.dr.xpos)))
        self.indx_R = int(np.argmin(np.abs(scale - self.parentFn.dr1.xpos)))

        # Run the fitting.
        res = minimize(self.residual, param, args=(
            scale[self.indx_L:self.indx_R+1], data[self.indx_L:self.indx_R+1], self.parentFn.def_options))

        # Repopulate the parameter object.
        for art in self.param:
            if self.param[art]['to_show'] == True and self.param[art]['calc'] == False:
                self.param[art]['value'] = param[art].value
                self.param[art]['stderr'] = float(param[art].stderr)

        # Extra parameters, which are not fit or inherited.
        # self.param['N_FCS']['value'] = np.round(1/self.param['GN0']['value'],4)

        self.residualVar = res.residual
        output = fit_report(param)
        print('residual', res.chisqr)
        if (res.chisqr > 0.05):
            print('CAUTION DATA DID NOT FIT WELL CHI^2 >0.05', res.chisqr)
            self.goodFit = False
        else:
            self.goodFit = True
        self.fitted = True
        self.chisqr = res.chisqr

        self.localTime = time.asctime(time.localtime(time.time()))

        # self.parentFn.updateTableFirst();
        self.model_autoNorm = equation_(
            param, scale[self.indx_L:self.indx_R+1], self.parentFn.def_options)
        self.model_autotime = scale[self.indx_L:self.indx_R+1]
        # self.parentFn.on_show()

        # self.parentFn.axes.plot(model_autotime,model_autoNorm, 'o-')
        # self.parentFn.canvas.draw();

    def load_from_file(self, channel):
        tscale = []
        tdata = []
        int_tscale = []
        int_tdata = []

        if self.ext == 'fcs':

            corrObj = self
            text = [0]
            r_obj = csv.reader(open(corrObj.filepath, 'rb'), delimiter='\t')
            title = r_obj.next()
            line = r_obj.next()
            line = r_obj.next()
            name = line[1].split(' = ')[1]
            read = True
            while read == True:

                corrObj = corrObject(self.filepath, self.parentFn)
                self.parentFn.objIdArr.append(corrObj)

                corrObj.name = name

                line = r_obj.next()
                text = []
                for part in line:
                    if part != '':
                        text.append(part)

                # Reads to first correlation array text.
                while text[0].split(' = ')[0] != 'CorrelationArray' or text[0].split(' = ')[1] == int(0):

                    line = r_obj.next()
                    text = []
                    for part in line:
                        if part != '':
                            text.append(part)
                line = r_obj.next()
                tdata = []
                tscale = []

                while text[0].split(' = ')[0] != 'PhotonCountHistogramArraySize':

                    try:
                        line = r_obj.next()
                    except:
                        read = False
                        break
                    text = []
                    for part in line:
                        if part != '':
                            text.append(part)
                    if text.__len__() > 1:
                        tscale.append(float(text[0]))
                        tdata.append(float(text[1]))

                if tdata.__len__() == 0:
                    corrObj = []
                    self.parentFn.objIdArr.pop(-1)
                    break

                channel = 0
                corrObj.siblings = None
                corrObj.autoNorm = np.array(
                    tdata).astype(np.float64).reshape(-1)
                corrObj.autotime = np.array(tscale).astype(
                    np.float64).reshape(-1)*1000
                corrObj.name = corrObj.name+'-CH'+str(channel)
                corrObj.ch_type = channel
                corrObj.param = copy.deepcopy(self.parentFn.def_param)
                self.parentFn.fill_series_list()

        if self.ext == 'SIN':
            self.parentFn.objIdArr.append(self.objId)
            proceed = False

            for line in csv.reader(open(self.filepath, 'rb'), delimiter='\t'):

                if proceed == 'correlated':
                    if line == []:
                        proceed = False
                    else:
                        tscale.append(float(line[0]))
                        tdata.append(float(line[channel+1]))
                if proceed == 'intensity':

                    if line == []:
                        proceed = False
                    elif line.__len__() > 1:

                        int_tscale.append(float(line[0]))
                        int_tdata.append(float(line[channel+1]))
                if (str(line) == "[\'[CorrelationFunction]\']"):
                    proceed = 'correlated'
                elif (str(line) == "[\'[IntensityHistory]\']"):
                    proceed = 'intensity'

            self.siblings = None
            self.autoNorm = np.array(tdata).astype(np.float64).reshape(-1)
            self.autotime = np.array(tscale).astype(
                np.float64).reshape(-1)*1000
            self.name = self.name+'-CH'+str(channel)
            self.ch_type = channel
            # self.prepare_for_fit()

            # Average counts per bin. For it to be seconds (Hz), must divide by duration.
            unit = int_tscale[-1]/(int_tscale.__len__()-1)
            # And to be in kHz we divide by 1000.
            self.kcount = np.average(np.array(int_tdata)/unit)/1000
            self.param = copy.deepcopy(self.parentFn.def_param)
            self.parentFn.fill_series_list()

            # Where we add the names.

        if self.ext == 'csv' or self.ext == 'CSV':
            r_obj = csv.reader(open(self.filepath, 'rb'))
            line_one = r_obj.next()
            if line_one.__len__() > 1:

                if float(line_one[1]) == 2:

                    version = 2
                else:
                    print('version not known:', line_one[1])

            else:
                version = 1

            if version == 1:
                self.parentFn.objIdArr.append(self)

                c = 0

                for line in csv.reader(open(self.filepath, 'rb')):
                    if (c > 0):
                        tscale.append(line[0])
                        tdata.append(line[1])
                    c += 1

                self.autoNorm = np.array(tdata).astype(np.float64).reshape(-1)
                self.autotime = np.array(tscale).astype(np.float64).reshape(-1)
                self.name = self.name+'-CH'+str(0)
                self.ch_type = 0
                self.datalen = len(tdata)

                self.param = copy.deepcopy(self.parentFn.def_param)
                self.parentFn.fill_series_list()
            if version >= 2:

                numOfCH = float(r_obj.next()[1])

                if numOfCH == 1:
                    self.parentFn.objIdArr.append(self)
                    self.type = str(r_obj.next()[1])
                    self.ch_type = int(r_obj.next()[1])
                    self.name = self.name+'-CH'+str(self.ch_type)

                    line = r_obj.next()

                    while line[0] != 'Time (ns)':
                        if line[0] == 'kcount':
                            self.kcount = float(line[1])
                        if line[0] == 'numberNandB':
                            self.numberNandB = float(line[1])
                        if line[0] == 'brightnessNandB':
                            self.brightnessNandB = float(line[1])
                        if line[0] == 'CV':
                            self.CV = float(line[1])
                        if line[0] == 'carpet pos':
                            carpet = int(line[1])
                        if line[0] == 'pc':
                            pc_text = int(line[1])
                        if line[0] == 'pbc_f0':
                            self.pbc_f0 = float(line[1])
                        if line[0] == 'pbc_tb':
                            self.pbc_tb = float(line[1])

                        line = r_obj.next()

                    if pc_text != False:
                        self.name = self.name + '_pc_m'+str(pc_text)

                    tscale = []
                    tdata = []
                    null = r_obj.next()
                    line = r_obj.next()
                    while line[0] != 'end':

                        tscale.append(line[0])
                        tdata.append(line[1])
                        line = r_obj.next()

                    self.autoNorm = np.array(tdata).astype(
                        np.float64).reshape(-1)
                    self.autotime = np.array(tscale).astype(
                        np.float64).reshape(-1)
                    self.siblings = None
                    self.param = copy.deepcopy(self.parentFn.def_param)
                    self.parentFn.fill_series_list()

                if numOfCH == 2:
                    corrObj2 = corrObject(self.filepath, self.parentFn)
                    corrObj3 = corrObject(self.filepath, self.parentFn)

                    self.parentFn.objIdArr.append(self)
                    self.parentFn.objIdArr.append(corrObj2)
                    self.parentFn.objIdArr.append(corrObj3)

                    line_type = r_obj.next()
                    self.type = str(line_type[1])
                    corrObj2.type = str(line_type[1])
                    corrObj3.type = str(line_type[1])

                    line_ch = r_obj.next()
                    self.ch_type = int(line_ch[1])
                    corrObj2.ch_type = int(line_ch[2])
                    corrObj3.ch_type = int(line_ch[3])

                    self.name = self.name+'-CH'+str(self.ch_type)
                    corrObj2.name = corrObj2.name+'-CH'+str(corrObj2.ch_type)
                    corrObj3.name = corrObj3.name+'-CH'+str(corrObj3.ch_type)

                    line = r_obj.next()
                    while line[0] != 'Time (ns)':
                        if line[0] == 'kcount':
                            self.kcount = float(line[1])
                            corrObj2.kcount = float(line[2])
                        if line[0] == 'numberNandB':
                            self.numberNandB = float(line[1])
                            corrObj2.numberNandB = float(line[2])
                        if line[0] == 'brightnessNandB':
                            self.brightnessNandB = float(line[1])
                            corrObj2.brightnessNandB = float(line[2])
                        if line[0] == 'CV':
                            self.CV = float(line[1])
                            corrObj2.CV = float(line[2])
                            corrObj3.CV = float(line[3])
                        if line[0] == 'carpet pos':
                            self.carpet_position = int(line[1])
                        if line[0] == 'pc':
                            pc_text = int(line[1])
                        if line[0] == 'pbc_f0':
                            self.pbc_f0 = float(line[1])
                            corrObj2.pbc_f0 = float(line[2])
                        if line[0] == 'pbc_tb':
                            self.pbc_tb = float(line[1])
                            corrObj2.pbc_tb = float(line[2])

                        line = r_obj.next()

                    if pc_text != False:
                        self.name = self.name + '_pc_m'+str(pc_text)
                        corrObj2.name = corrObj2.name + '_pc_m'+str(pc_text)
                        corrObj3.name = corrObj3.name + '_pc_m'+str(pc_text)

                    null = r_obj.next()
                    line = r_obj.next()
                    tscale = []
                    tdata0 = []
                    tdata1 = []
                    tdata2 = []
                    while line[0] != 'end':

                        tscale.append(line[0])
                        tdata0.append(line[1])
                        tdata1.append(line[2])
                        tdata2.append(line[3])
                        line = r_obj.next()

                    self.autotime = np.array(tscale).astype(
                        np.float64).reshape(-1)
                    corrObj2.autotime = np.array(
                        tscale).astype(np.float64).reshape(-1)
                    corrObj3.autotime = np.array(
                        tscale).astype(np.float64).reshape(-1)

                    self.autoNorm = np.array(tdata0).astype(
                        np.float64).reshape(-1)
                    corrObj2.autoNorm = np.array(
                        tdata1).astype(np.float64).reshape(-1)
                    corrObj3.autoNorm = np.array(
                        tdata2).astype(np.float64).reshape(-1)

                    self.siblings = [corrObj2, corrObj3]
                    corrObj2.siblings = [self, corrObj3]
                    corrObj3.siblings = [self, corrObj2]

                    self.param = copy.deepcopy(self.parentFn.def_param)
                    corrObj2.param = copy.deepcopy(self.parentFn.def_param)
                    corrObj3.param = copy.deepcopy(self.parentFn.def_param)
                    self.parentFn.fill_series_list()
